"""The Generated Driving Grid World Environment."""
from typing import Any, Dict, Optional, Set, Tuple, Union

import posggym.model as M
from posggym.envs.grid_world.core import (
    Coord,
    GridCycler,
    GridGenerator,
)
from posggym.envs.grid_world.driving import DrivingEnv, DrivingGrid, DObs


class DrivingGenEnv(DrivingEnv):
    """The Generated Driving Grid World Environment.

    This is the same as the [Driving](/environments/grid_world/driving) Environment
    except that a new grid is generated at each reset.

    For environment attributes see [Driving](/environments/grid_world/driving)
    environment class documentation.

    Arguments
    ---------

    - `num_agents` - the number of agents in the environment (default = `2`).
    - `obs_dim` - the local observation dimensions, specifying how many cells in front,
        behind, and to each side the agent observes (default = `(3, 1, 1)`, resulting
        in the agent observing a 5x3 area: 3 in front, 1 behind, 1 to each side.)
    - `obstacle_collisions` -  whether running into a wall results in the agent's
        vehicle crashing and thus the agent reaching a terminal state. This can make
        the problem significantly harder (default = "False").
    - `generator_params` - the parameters to use for generating the grid
        (default = "14x14"). This can either be a string specyfing one of the supported
        sets of params (can be "7x7", "14x14", "28x28") or a dictionary with the
        following keys:
        - `width` - width of the grid
        - `height` - height of the grid
        - `max_obstacle_size` - max size of an obstacle
        - `max_num_obstacles` - maximum number of obstacles in the grid
    - `n_grids` - the number of different grids to generate, if provided then `n_grids`
        will be generated and these will be cycled through for each reset. If `None`
        then a different, possibly unique, grid will be generated each episode
        (default = `None`).
    - `shuffle_grid_order` - whether to shuffle the order in which different grids
        appear. Only has any effect if `n_grids > 1` (default = `True`).

    Available variants
    ------------------

    The DrivingGen environment comes with a number of pre-built sets of grid generator
    parameters which can be passed as an argument to `posggym.make`:

    | Name    | Grid size | max obstacle size | max num obstacles |
    |---------|-----------|-------------------|-------------------|
    | `7x7`   | 7x7       | 2                 | 21                |
    | `14x14` | 14x14     | 3                 | 42                |
    | `28x28` | 28x28     | 4                 | 84                |


    For example to use the DrivingGen environment with the pre-build `7x7` set of
    generation parameters and 2 agents, you would use:

    ```python
    import posggym
    env = posgggym.make('DrivingGen-v0', generator_params="7x7", num_agents="2")
    ```

    Version History
    ---------------
    - `v0`: Initial version

    """

    def __init__(
        self,
        num_agents: int = 2,
        obs_dim: Tuple[int, int, int] = (3, 1, 2),
        obstacle_collisions: bool = False,
        generator_params: Union[str, Dict[str, int]] = "14x14",
        n_grids: Optional[int] = None,
        shuffle_grid_order: bool = True,
        render_mode: Optional[str] = None,
    ):
        if isinstance(generator_params, str):
            assert generator_params in SUPPORTED_GEN_PARAMS, (
                f"Unsupported grid generator parameters'{generator_params}'. If "
                "grid `generator_params` argument is a string it must be one of: "
                f"{SUPPORTED_GEN_PARAMS.keys()}."
            )
            generator_params = SUPPORTED_GEN_PARAMS[generator_params][0]

        self._generator_params = generator_params
        self._n_grids = n_grids
        self._shuffle_grid_order = shuffle_grid_order
        self._gen = DrivingGridGenerator(**self._generator_params)

        if n_grids is not None:
            grids = self._gen.generate_n(n_grids)
            self._cycler = GridCycler(grids, shuffle_grid_order)
            grid: "DrivingGrid" = grids[0]  # type: ignore
        else:
            self._cycler = None  # type: ignore
            grid = self._gen.generate()

        super().__init__(
            grid,
            num_agents,
            obs_dim,
            obstacle_collisions=obstacle_collisions,
            render_mode=render_mode,
        )

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[M.AgentID, DObs], Dict[M.AgentID, Dict]]:
        if seed is not None:
            self._model_seed = seed
            self._gen = DrivingGridGenerator(seed=seed, **self._generator_params)

            if self._n_grids is not None:
                grids = self._gen.generate_n(self._n_grids)
                self._cycler = GridCycler(grids, self._shuffle_grid_order, seed=seed)

        grid = self._cycler.next() if self._n_grids else self._gen.generate()

        self.model.grid = grid  # type: ignore

        if self.render_mode != "ansi" and self.renderer is not None:
            self.renderer.grid = grid
            self.renderer.reset_blocks()

        return super().reset(seed=seed)


class DrivingGridGenerator(GridGenerator):
    """Class for generating grid layouts for Driving Environment.

    Generates starting and destination coords in an alternating pattern along
    the outside of edge of the grid.
    """

    def __init__(
        self,
        width: int,
        height: int,
        max_obstacle_size: int,
        max_num_obstacles: int,
        seed: Optional[int] = None,
    ):
        super().__init__(
            width,
            height,
            self._generate_mask(width, height),
            max_obstacle_size,
            max_num_obstacles,
            ensure_grid_connected=True,
            seed=seed,
        )

        self._start_coords = [self.mask for _ in range(len(self.mask))]
        self._dest_coords = [self.mask for _ in range(len(self.mask))]

    def _generate_mask(self, width: int, height: int) -> Set[Coord]:
        start = 1
        mask = set()
        for x in range(start, width, 2):
            mask.add((x, 0))
            mask.add((x, height - 1))

        for y in range(start, height, 2):
            mask.add((0, y))
            mask.add((width - 1, y))

        return mask

    def generate(self) -> DrivingGrid:
        """Generate a new Driving grid."""
        base_grid = super().generate()
        driving_grid = DrivingGrid(
            base_grid.width,
            base_grid.height,
            base_grid.block_coords,
            self._start_coords,
            self._dest_coords,
        )
        return driving_grid


# (generator params, finite horizon step limit)
SUPPORTED_GEN_PARAMS = {
    "7x7": (
        {
            "width": 7,
            "height": 7,
            "max_obstacle_size": 2,
            "max_num_obstacles": 21,  # size * 3
        },
        50,
    ),
    "14x14": (
        {
            "width": 14,
            "height": 14,
            "max_obstacle_size": 3,
            "max_num_obstacles": 42,  # size * 3
        },
        50,
    ),
    "28x28": (
        {
            "width": 28,
            "height": 28,
            "max_obstacle_size": 4,
            "max_num_obstacles": 84,  # size * 3
        },
        100,
    ),
}



"""The model data structure."""
from __future__ import annotations

import abc
import dataclasses
import enum
import random
from typing import (
    TYPE_CHECKING,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
)

import numpy as np

from posggym import error
from posggym.utils import seeding


if TYPE_CHECKING:
    from gymnasium import spaces
    from posggym.envs.registration import EnvSpec


AgentID = str
StateType = TypeVar("StateType")
ActType = TypeVar("ActType")
ObsType = TypeVar("ObsType")


@dataclasses.dataclass(order=True)
class JointTimestep(Generic[StateType, ObsType]):
    """Stores values returned by model after a single step.

    Supports iteration.

    A dataclass is used instead of a Namedtuple so that generic typing is seamlessly
    supported.

    """

    state: StateType
    observations: Dict[AgentID, ObsType]
    rewards: Dict[AgentID, float]
    terminated: Dict[AgentID, bool]
    truncated: Dict[AgentID, bool]
    all_done: bool
    info: Dict[AgentID, Dict]

    def __iter__(self):
        for field in dataclasses.fields(self):
            yield getattr(self, field.name)


class POSGModel(abc.ABC, Generic[StateType, ObsType, ActType]):
    r"""A Partially Observable Stochastic Game model.

    This class defines functions and attributes necessary for a generative POSG
    model for use in simulation-based planners (e.g. MCTS), and for reinforcement
    learning.

    The main API methods that users of this class need to know are:

    - :meth:`get_agents` - Returns the IDs of agents currently active for a given state
    - :meth:`sample_initial_state` - Returns an a sampled initial state of the
      environment.
    - :meth:`sample_initial_obs` - Returns a sampled observation given an initial state.
    - :meth:`step` - The generative step function which given a state and agent actions,
      returns the next state, agent observations, agent rewards, whether the environment
      has terminated or truncated for each agent, whether the environment has reached
      a terminal state for all agents, and information from the environment about the
      step
    - :meth:`seed` - Sets the seed for the models random number generator.

    Additionally, the main API attributes are:

    - :attr:`possible_agents` - All possible agents that may appear in the environment
      across all states.
    - :attr:`state_space` - The space of all environment states. Noting that an explicit
      state space definition is not needed by many simulation-based algorithms
      (including RL and MCTS) and can be hard to define so implementing this property
      should be seen as optional. In cases where it is not implemented it should be
      None.
    - :attr:`action_spaces` - The action space for each agent
    - :attr:`observations_spaces` - The observation space for each agent
    - :attr:`reward_ranges` - The minimum and maximum possible rewards over an episode
      for each agent. The default reward range is set to :math:`(-\infty,+\infty)`.
    - :attr:`is_symmetric` - Whether the environment is symmetric or asymmetric. That is
      whether all agents are identical irrespective of their ID (i.e. same actions,
      observation, and reward spaces and dynamics)
    - :attr:`spec` - An environment spec that contains the information used to
      initialize the environment from :meth:`posggym.make`
    - :attr:`rng` - the model's internal random number generator (RNG).

    Custom environments should inherit from this class and implement the
    :meth:`get_agents`, :meth:`sample_initial_state`, :meth:`sample_initial_obs`,
    :meth:`step` methods and the :attr:`possible_agents`, :attr:`action_spaces`,
    :attr:`observations_spaces`, :attr:`is_symmetric`, :attr:`rng` attributes.

    Custom environments may optionally provide implementations for the
    :meth:`sample_agent_initial_state` method and :attr:`state_space` attribute.

    Note
    ----
    The POSGGym Model API models all environments as environments that are
    `observation first`, that is the environment provides an initial observation before
    any action is taken (rather than action first, where agents perform an action
    before any observation is received). `observation first` environments are the
    standard in reinforcement learning problems and also for most real world problems,
    and are becoming the more common model API. It's also trivial to convert an
    `action first` model into `observation first` by just returning a default or dummy
    initial observation (e.g. the initial observation is always the first observation
    in the list of possible observations).

    """

    # EnvSpec used to instantiate env instance this model is for
    # This is set when env is made using posggym.make function
    spec: Optional["EnvSpec"] = None

    # All agents that may appear in the environment
    possible_agents: Tuple[AgentID, ...]
    # State space
    state_space: spaces.Space | None = None
    # Action space for each agent
    action_spaces: Dict[AgentID, spaces.Space]
    # Observation space for each agent
    observation_spaces: Dict[AgentID, spaces.Space]
    # Whether the environment is symmetric or not (is asymmetric)
    is_symmetric: bool
    # Random number generator, created as needed by `rng` method.
    _rng: seeding.RNG | None = None

    @property
    def reward_ranges(self) -> Dict[AgentID, Tuple[float, float]]:
        r"""A mapping from Agent ID to min and max possible rewards for that agent.

        Each reward tuple corresponding to the minimum and maximum possible rewards for
        a given agent over an episode. The default reward range for each agent is set to
        :math:`(-\infty,+\infty)`.

        Returns
        -------
        Dict[AgentID, Tuple[float, float]]

        """
        return {i: (-float("inf"), float("inf")) for i in self.possible_agents}

    @abc.abstractmethod
    def get_agents(self, state: StateType) -> List[AgentID]:
        """Get list of IDs for all agents that are active in given state.

        These list of active agents may change depending on state.

        For any environment where the number of agents remains constant during AND
        across episodes. This will be :attr:`possible_agents`, independent of state.

        Returns
        -------
        List[AgentID]
          List of IDs for all agents that active in given state,

        """

    @abc.abstractmethod
    def sample_initial_state(self) -> StateType:
        """Sample an initial state.

        Returns
        -------
        StateType
          An initial state.

        """

    @abc.abstractmethod
    def sample_initial_obs(self, state: StateType) -> Dict[AgentID, ObsType]:
        """Sample initial agent observations given an initial state.

        Arguments
        ---------
        state : StateType
          The initial state.

        Returns
        -------
        Dict[AgentID, ObsType]
          A mapping from AgentID to initial observation.

        """

    @abc.abstractmethod
    def step(
        self, state: StateType, actions: Dict[AgentID, ActType]
    ) -> JointTimestep[StateType, ObsType]:
        """Perform generative step.

        For custom environments that have win/loss or success/fail conditions, you are
        encouraged to include this information in the `info` property of the returned
        value. We suggest using the the "outcomes" key with an instance of the
        ``Outcome`` class for values.

        """

    @property
    @abc.abstractmethod
    def rng(self) -> seeding.RNG:
        """Return the model's internal random number generator (RNG).

        Initializes RNG with a random seed if not yet initialized.

        `posggym` models and environments support the use of both the python built-in
        random library and the numpy library, unlike `gymnasium` which only explicitly
        supports the numpy library. Support for the built-in library is included as it
        can be 2-3X faster than the numpy library when drawing single samples, providing
        a significant speed-up for many environments.

        There's also nothing stopping users from using other RNG libraries so long as
        they implement the model API. However, explicit support in the form of tests
        and type hints is only provided for the random and numpy libraries.

        Returns
        -------
        rng : model's internal random number generator.

        """

    def seed(self, seed: Optional[int] = None):
        """Set the seed for the model RNG.

        Also handles seeding for the action, observation, and (if it exists) state
        spaces.

        """
        if isinstance(self.rng, random.Random):
            self._rng, seed = seeding.std_random(seed)
        elif isinstance(self.rng, np.random.Generator):
            self._rng, seed = seeding.np_random(seed)
        else:
            raise error.UnseedableEnv(
                "{self.__class__.__name__} unseedable. Please ensure the model has "
                "implemented the rng property. The model class must also overwrite "
                "the `seed` method if it uses a RNG not from the `random` or "
                "`numpy.random` libraries."
            )

        seed += 1
        for act_space in self.action_spaces.values():
            act_space.seed(seed)
            seed += 1

        for obs_space in self.observation_spaces.values():
            obs_space.seed(seed)
            seed += 1

        if self.state_space is not None:
            self.state_space.seed(seed)

    def sample_agent_initial_state(self, agent_id: AgentID, obs: ObsType) -> StateType:
        """Sample an initial state for an agent given it's initial observation.

        It is optional to implement this method but can helpful in environments that are
        used for planning and where there are a huge number of possible initial states.

        Arguments
        ---------
        agent_id : Union[int, str]
          The ID of the agent to get initial state for.
        obs : ObsType
          The initial observation of the agent.

        Returns
        -------
        StateType
          An initial state for the agent conditioned on their initial observation.

        Raises
        ------
        NotImplementedError
          If this method is not implemented.

        """
        raise NotImplementedError


class POSGFullModel(POSGModel[StateType, ObsType, ActType], abc.ABC):
    r"""A Fully defined Partially Observable Stochastic Game model.

    This class inherits from the :py:class:`POSGModel`, adding implementions for all
    individual components of a POSG. It is designed for use by planners which utilize
    all model components (e.g. dynamic programming and full-width planners).

    The additional methods that need to be implemented are:

    - :meth:`get_initial_belief` - the initial belief distribution :math:`b_{0}`,
      mapping initial states to probabilities.
    - :meth:`transition_fn` - the transition function
      :math:`T(s, a, s') \rightarrow [0, 1]` defining :math:`Pr(s'|s, a)`, the
      probability of getting next state `s'` given the environment was in state `s` and
      joint action `a` was performed.
    - :meth:`observation_fn` - the observation function
      :math:`Z(o, s', a) \rightarrow [0, 1]` defining :math:`Pr(o|s', a)`, the
      probability of joint observation `o` given the joint action `a` was performed and
      the environment ended up in state `s'`
    - :meth:`reward_fn` : the reward function
      :math:`R(s, a) \rightarrow \mathbf{R}^n` where `n` is the number of agents,
      defines the reward each agent receives given joint action `a` was performed in
      state `s`.


    """

    @abc.abstractmethod
    def get_initial_belief(self) -> Dict[StateType, float]:
        """The initial belief distribution: :math:`b_{0}`.

        Returns
        -------
        Dict[StateType, float]
           :math:`Pr(s_{0}=s)` the initial probability of each state. If a state is not
           included in the initial distribution object, it should be assumed to have
           probability 0.

        """

    @abc.abstractmethod
    def transition_fn(
        self, state: StateType, actions: Dict[AgentID, ActType], next_state: StateType
    ) -> float:
        """Transition function :math:`T(s', a, s)`.

        Arguments
        ---------
        state : StateType
          the state the environment was in
        actions : Dict[AgentID, ActType]
          the joint action performed
        next_state : StateType
          the state of the environment after actions were performed

        Returns
        -------
        float
          :math:`Pr(s'|s, a)`, the probability of getting next state `s'` given the
          environment was in state `s` and joint action `a` was performed.

        """

    @abc.abstractmethod
    def observation_fn(
        self,
        obs: Dict[AgentID, ObsType],
        next_state: StateType,
        actions: Dict[AgentID, ActType],
    ) -> float:
        """Observation function :math:`Z(o, s', a)`.

        Arguments
        ---------
        obs : Dict[AgentID, ObsType]
          the observation received
        actions : Dict[AgentID, ActType]
          the joint action performed
        next_state : StateType
          the state of the environment after actions were performed

        Returns
        -------
        float
          :math:`Pr(o|s', a)`, the probability of joint observation `o` given the joint
          action `a` was performed and the environment ended up in state `s'`.

        """

    @abc.abstractmethod
    def reward_fn(
        self, state: StateType, actions: Dict[AgentID, ActType]
    ) -> Dict[AgentID, float]:
        """The reward Function :math:`R(s, a)`.

        Arguments
        ---------
        state : StateType
          the state the environment was in
        actions : Dict[AgentID, ActType]
          the joint action performed

        Returns
        -------
        Dict[AgentID, float]
          The reward each agent receives given joint action `a` was performed in state
          `s`.

        """


class Outcome(enum.Enum):
    """An enum for final episode Outcome of an agent.

    This is supplied for user convenience. For environments where agents can WIN/LOSE,
    this class can be used to supply that information to users in a standard format via
    the ``info`` return value of the :meth:`POSGModel.step()` function.

    Has the following possible values:

    - :attr:`LOSS` = -1
    - :attr:`DRAW` = 0
    - :attr:`WIN` = 1
    - :attr:`NA` = None

    """

    LOSS = -1
    DRAW = 0
    WIN = 1
    NA = None

    def __str__(self):
        return self.name



"""The Level-Based Foraging Environment."""
import enum
import math
from collections import defaultdict
from itertools import product
from os import path
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

import numpy as np
from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.utils import seeding


class Player(NamedTuple):
    """A player in the Level-Based Foraging environment."""

    idx: int
    coord: Coord
    level: int


class Food(NamedTuple):
    """A food in the Level-Based Foraging environment."""

    coord: Coord
    level: int


class LBFState(NamedTuple):
    """State in Level-Based Foraging environment."""

    players: Tuple[Player, ...]
    food: Tuple[Food, ...]
    food_spawned: int


class LBFAction(enum.IntEnum):
    """Level-Based Foraging Actions."""

    NONE = 0
    NORTH = 1
    SOUTH = 2
    WEST = 3
    EAST = 4
    LOAD = 5


ACTION_TO_DIR = {
    LBFAction.NORTH: Direction.NORTH,
    LBFAction.SOUTH: Direction.SOUTH,
    LBFAction.WEST: Direction.WEST,
    LBFAction.EAST: Direction.EAST,
}


class CellEntity(enum.IntEnum):
    """Entity encodings for grid observations."""

    OUT_OF_BOUNDS = 0
    EMPTY = 1
    FOOD = 2
    AGENT = 3


LBFObs = Union[Tuple[int, ...], np.ndarray]


class LBFEntityObs(NamedTuple):
    """Entity (player, food) observation in the LBF environment."""

    coord: Coord
    level: int
    is_self: bool


class LBFEnv(DefaultEnv[LBFState, LBFObs, LBFAction]):
    """The Level-Based Foraging Environment.

    This implementation is based on the original implementation of Level-Based Foraging
    environment: <https://github.com/semitable/lb-foraging>. We modify their original
    version to support access to the environments dynamic's model as well as adding
    more control over the layout of food in the world.

    The Level-Based Foraging is a 2D grid-world involving multiple agents each of which
    is trying to collect as much food as possible. A finite amount of food is spread
    throughout the world which the agents can collect. The key interesting feature is
    that both the agents and the food have levels and a piece of food can only be picked
    up if the sum of the levels of all the agents trying to pick up the food is greater
    than the foods level. This incentivizes cooperation between agents.

    The problem can be set in two-modes: `cooperative` and `mixed`. In cooperative mode
    agents share rewards and so are fully incentivized to work together. While in mixed
    mode agents rewards are given individually so they are incentivized to cooperate to
    collect food with higher levels, but also incentivized to act greedily to collect
    lower level food by themselves, creating an interesting social-dilemma.

    Agents
    ------
    Between 2 and 4, with all agents active throughout every episode.

    State Space
    -----------
    The state of the environment is defined by a `(x, y, level)` triplet for each agent
    and food object in the environment. The `(x, y)` components define the position of
    the agent or food, starting from the bottom left square. The `level` component is
    the level of the agent or food.

    Action Space
    ------------
    Each agent has six possible discrete actions: `NOOP=0`, `NORTH=1`, `SOUTH=2`,
    `WEST=3`, `EAST=4`, and `LOAD=5`. The NORTH, SOUTH, WEST, EAST actions move the
    agent in the given direction, while the LOAD action attempts to pickup any adjacent
    food. The NOOP action does nothing.

    Observation Space
    -----------------
    Each agent observes the `(x, y, level)` of food and other agent within their field
    of vision, which is `sight` distance away from the agent in all directions.

    There are three observation modes:

    1. *grid*
        - The agent receives three 2D layers of size (`1+2*sight`, `1+2*sight`). Each
          cell in each layer corresponds to a specific (x, y) coordinate relative to
          the observing agent. The layers are:
            1. agent level
            2. food level
            3. whether cell is free or blocked (e.g. out of bounds)
    2. *vector*
        - A vector of `(x, y, level)` triplets for each food and agent in the
          environment. If a given food or agent is not within the observing agent's
          field of vision triplets have a value of `(-1, -1, 0)`. The size of the vector
          is `(num_agents + max_food) * 3`, with the first `max_food` triplets being
          for the food, the `max_food+1` triplet being for the observing agent and the
          remaining `num_agents-1` triplets for the other agents. The ordering of the
          triplets for the other agents is consistent, while the food obs triplets can
          change based on how many are visible and their relative coordinates to the
          observing agent.
    3. *tuple*
       - This is the same as the vector observation except observations are Python
         tuples of integers instead of numpy arrays of floats.

    Rewards
    -------
    Agents receive a reward whenever they successfully pick-up food. The reward received
    by each agent depends on the food's level and how many agents picked up the food and
    their levels. In cooperative mode all agents receive the same reward.

    Dynamics
    --------
    Actions are deterministic with each agent's movement action resulting in them moving
    1 cell in the given direction so long as that cell is not occupied or out-of-bounds.
    The `LOAD` action will only succeed if the agent is adjacent to a food object and
    the agents level is higher than the food level, or other agents are attempting to
    load the food at the same time, and the sum of the all loading agent's levels is
    greater than the food's level. If a food is successfully loaded then it is removed
    from the map.

    Starting State
    --------------
    The initial state depends on if the environment is using a `static_layout` or not.

    When using a static layout each agent will start from the same location, in one of
    the corners or half-way along one of the grid's edges, each episodes and the food
    will be in the same positions. The level of each agent and food will be selected
    randomly. The static layout is useful when using a planner since the number of
    possible initial states is significantly reduced.

    When not using the static layout, agent starting and food locations are selected
    randomly from all possible positions, and the agent and food levels are also
    selected randomly.

    Episodes End
    ------------
    Episodes end when all food has been collected. By default a `max_episode_steps` is
    also set for each Driving environment. The default value is `50` steps, but this may
    need to be adjusted when using larger grids (this can be done by manually specifying
    a value for `max_episode_steps` when creating the environment with `posggym.make`).

    Arguments
    ---------

    - `num_agents` - the number of agents in the environment (default = `2`).
    - `max_agent_level` - the maximum level of an agent (default = `3`).
    - `field_size` - the width and height of the grid world (default = `(10, 10)`).
    - `max_food` - the maximum number of food that will appear in an episode
        (default = `8`).
    - `sight` - the local observation dimensions, specifying how many cells in each
        direction the agent observes (default = `2`, resulting in the agent observing a
        5x5 area)
    - `force_coop` -  whether all agents share all rewards (i.e. fully cooperative mode)
        (default = "False").
    - `static_layout` - whether to use a static food layout. If true then the same
        number of food will always appear each episode and always in the same locations.
        The level of the food will be random each episode. If false, food location and
        levels is random each episode (default = 'False`)
    - `normalize_reward` - whether to normalize rewards so that the max sum of rewards
        for an episode is 1.0 (default=`True`).
    - `observation_mode` - the observation mode to use out of `grid`, `vector`, or
        `tuple` (default=`tuple`)
    - `penalty` - the penalty for failing to load food (default=0.0)

    Version History
    ---------------
    - `v2`: Version adapted from <https://github.com/semitable/lb-foraging>

    References
    ----------
    - Stefano V. Albrecht and Subramanian Ramamoorthy. 2013. A Game-Theoretic Model and
      Best-Response Learning Method for Ad Hoc Coordination in Multia-gent Systems.
      In Proceedings of the 2013 International Conference on Autonomous Agents and
      Multi-Agent Systems. 1155–1156.
    - S. V. Albrecht and Peter Stone. 2017. Reasoning about Hypothetical Agent
      Behaviours and Their Parameters. In 16th International Conference on Autonomous
      Agents and Multiagent Systems 2017. International Foundation for Autonomous
      Agents and Multiagent Systems, 547–555
    - Filippos Christianos, Lukas Schäfer, and Stefano Albrecht. 2020. Shared Experience
      Actor-Critic for Multi-Agent Reinforcement Learning. Advances in Neural
      Information Processing Systems 33 (2020), 10707–10717
    - Georgios Papoudakis, Filippos Christianos, Lukas Schäfer, and Stefano V. Albrecht.
      2021. Benchmarking Multi-Agent Deep Reinforcement Learning Algorithms in
      Cooperative Tasks. In Thirty-Fifth Conference on Neural Information Processing
      Systems Datasets and Benchmarks Track (Round 1)

    """

    metadata = {
        "render_modes": ["human", "rgb_array", "rgb_array_dict"],
        "render_fps": 15,
    }

    def __init__(
        self,
        num_agents: int = 2,
        max_agent_level: int = 3,
        field_size: Tuple[int, int] = (10, 10),
        max_food: int = 8,
        sight: int = 2,
        force_coop: bool = False,
        static_layout: bool = False,
        normalize_reward: bool = True,
        observation_mode: str = "tuple",
        penalty: float = 0.0,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            LBFModel(
                num_agents,
                max_agent_level,
                field_size=field_size,
                max_food=max_food,
                sight=sight,
                force_coop=force_coop,
                static_layout=static_layout,
                normalize_reward=normalize_reward,
                observation_mode=observation_mode,
                penalty=penalty,
                **kwargs,
            ),
            render_mode=render_mode,
        )
        self.max_food = max_food
        self.renderer = None
        self.food_imgs = None
        self.agent_imgs = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return

        import posggym.envs.grid_world.render as render_lib

        model: LBFModel = self.model  # type: ignore
        if self.renderer is None:
            self.renderer = render_lib.GWRenderer(
                self.render_mode,
                model.grid,
                render_fps=self.metadata["render_fps"],
                env_name="LevelBasedForaging",
            )

        if self.agent_imgs is None:
            img_path = path.join(path.dirname(__file__), "img", "robot.png")
            agent_img = render_lib.load_img_file(img_path, self.renderer.cell_size)
            font = render_lib.load_font(
                "Comic Sans MC",
                render_lib.get_default_font_size(self.renderer.cell_size),
            )
            self.agent_imgs = {
                i: render_lib.GWImageAndText(
                    (0, 0), self.renderer.cell_size, agent_img, str(1), font
                )
                for i in self.possible_agents
            }

        if self.food_imgs is None:
            img_path = path.join(path.dirname(__file__), "img", "apple.png")
            food_img = render_lib.load_img_file(img_path, self.renderer.cell_size)
            font = render_lib.load_font(
                "Comic Sans MC",
                render_lib.get_default_font_size(self.renderer.cell_size),
            )
            self.food_imgs = [
                render_lib.GWImageAndText(
                    (0, 0), self.renderer.cell_size, food_img, str(1), font
                )
                for i in range(self.max_food)
            ]

        render_objects = []
        observed_coords = []
        for i, player in enumerate(self._state.players):
            img_obj = self.agent_imgs[str(i)]
            img_obj.coord = player.coord
            img_obj.text = str(player.level)
            render_objects.append(img_obj)
            observed_coords.extend(model.get_obs_coords(player.coord))

        for i, food in enumerate(self._state.food):
            img_obj = self.food_imgs[i]
            img_obj.coord = food.coord
            img_obj.text = str(food.level)
            render_objects.append(img_obj)

        agent_coords_and_dirs = {
            str(i): (player.coord, Direction.NORTH)
            for i, player in enumerate(self._state.players)
        }

        if self.render_mode in ("human", "rgb_array"):
            return self.renderer.render(render_objects, observed_coords)
        return self.renderer.render_agents(
            render_objects,
            agent_coords_and_dirs,
            agent_obs_dims=model.sight,
            observed_coords=observed_coords,
        )

    def close(self):
        if self.renderer is not None:
            self.renderer.close()
            self.renderer = None


class LBFModel(M.POSGModel[LBFState, LBFObs, LBFAction]):
    """Level-Based Foraging model.

    Parameters
    ----------
    num_agents : int
        the number of agents in the environment
    max_agent_level : int
        maximum foraging level of any agent
    field_size : Tuple[int, int]
        width and height of the playing field
    max_food : int
        the max number of food that can be spawned
    sight : int
        observation size of each agent
    force_coop : bool
        whether to force cooperation or not
    normalize_reward : bool, optional
        whether to normalize the reward or not (default=True)
    observation_mode : str, optional
        The observation mode for agent (default='tuple')
          - 'grid' - observations are multiple 2D grids (3D np.ndarray)
          - 'vector' - observations are vector (1D np.ndarray)
          - 'tuple' - observations are a tuple with same format as 'vector'
                      observations but as a hashable Python tuple object
                      containing integers instead of floats
    penalty : float, optional
        the penalty for failing to load food (default=0.0)

    """

    OBSERVATION_MODES = ["grid", "vector", "tuple"]

    # pylint: disable=unused-argument
    def __init__(
        self,
        num_agents: int,
        max_agent_level: int,
        field_size: Tuple[int, int],
        max_food: int,
        sight: int,
        force_coop: bool,
        static_layout: bool,
        normalize_reward: bool = True,
        observation_mode: str = "tuple",
        penalty: float = 0.0,
        **kwargs,
    ):
        assert observation_mode in self.OBSERVATION_MODES
        self.field_size = field_size
        self._rows, self._cols = field_size
        self._max_agent_level = max_agent_level
        self._max_food = max_food
        self.sight = sight
        self._force_coop = force_coop
        self._static_layout = static_layout
        self._normalize_reward = normalize_reward
        self.observation_mode = observation_mode
        self._penalty = penalty

        self.possible_agents = tuple(str(i) for i in range(num_agents))
        self.action_spaces = {
            i: spaces.Discrete(len(LBFAction)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: self.get_agent_observation_space() for i in self.possible_agents
        }
        self.is_symmetric = True

        self.grid = Grid(
            grid_width=self._cols, grid_height=self._rows, block_coords=None
        )

        self._food_locations: Optional[List[Coord]] = None
        self._player_locations: Optional[List[Coord]] = None
        if static_layout:
            assert (
                self._rows == self._cols
            ), "static layout only supported for square grids"
            self._food_locations = self._generate_static_food_coords()
            self._player_locations = self._generate_static_player_coords()

    def get_agent_observation_space(self) -> spaces.Space:
        """The Observation Space for an agent.

        - all of the board (board_size^2) with foods
        - player description (x, y, level)*player_count

        """
        n_agents = len(self.possible_agents)
        max_food_level = self._max_agent_level * min(n_agents, 4)
        if self.observation_mode == "tuple":
            max_x, max_y = self._cols, self._rows
            tuple_obs_space = []
            for _ in range(n_agents):
                tuple_obs_space.append(spaces.Discrete(max_x + 1, start=-1))
                tuple_obs_space.append(spaces.Discrete(max_y + 1, start=-1))
                tuple_obs_space.append(
                    spaces.Discrete(self._max_agent_level + 1, start=0)
                )
            for _ in range(self._max_food):
                tuple_obs_space.append(spaces.Discrete(max_x + 2, start=-1))
                tuple_obs_space.append(spaces.Discrete(max_y + 2, start=-1))
                tuple_obs_space.append(spaces.Discrete(max_food_level + 1, start=0))
            return spaces.Tuple(tuple_obs_space)

        if self.observation_mode == "vector":
            max_x, max_y = self._cols - 1, self._rows - 1
            min_obs = [-1, -1, 0] * n_agents + [-1, -1, 0] * self._max_food
            max_obs = [max_x, max_y, self._max_agent_level] * n_agents + [
                max_x,
                max_y,
                max_food_level,
            ] * self._max_food
        else:
            # grid observation space
            grid_shape = (1 + 2 * self.sight, 1 + 2 * self.sight)

            # agents layer: agent levels
            agents_min = np.zeros(grid_shape, dtype=np.float32)
            agents_max = np.ones(grid_shape, dtype=np.float32) * self._max_agent_level

            # foods layer: foods level
            foods_min = np.zeros(grid_shape, dtype=np.float32)
            foods_max = np.ones(grid_shape, dtype=np.float32) * max_food_level

            # access layer: i the cell available
            access_min = np.zeros(grid_shape, dtype=np.float32)
            access_max = np.ones(grid_shape, dtype=np.float32)

            # total layer
            min_obs = np.stack([agents_min, foods_min, access_min])
            max_obs = np.stack([agents_max, foods_max, access_max])

        # dtype = np.int8 if self.observation_mode == "vector" else np.uint8
        dtype = np.float32
        return spaces.Box(
            np.array(min_obs, dtype=dtype),
            np.array(max_obs, dtype=dtype),
            dtype=dtype,
        )

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[float, float]]:
        if self._normalize_reward:
            reward_range = (self._penalty, 1.0)
        else:
            reward_range = (self._penalty, self._max_agent_level)
        return {i: reward_range for i in self.possible_agents}

    def get_agents(self, state: LBFState) -> List[M.AgentID]:
        return list(self.possible_agents)

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def seed(self, seed: Optional[int] = None):
        super().seed(seed)
        if self._static_layout:
            assert (
                self._rows == self._cols
            ), "static layout only supported for square grids"
            self._food_locations = self._generate_static_food_coords()
            self._player_locations = self._generate_static_player_coords()

    def _generate_static_food_coords(self) -> List[Coord]:
        """Generate food coords for static layout.

        Number and location of food is the same for given pairing of
        (field size, max_food).
        """
        assert self._rows == self._cols
        # must be enough space to surround each food with free cells
        assert self._max_food <= math.floor((self._rows - 1) / 2) ** 2

        food_grid_size = math.floor((self._rows - 1) / 2)
        available_food_locs = sorted_from_middle(
            list(product(range(food_grid_size), repeat=2))
        )

        food_locs = []
        for f in range(self._max_food):
            row = available_food_locs[f][0] * 2 + 1
            col = available_food_locs[f][1] * 2 + 1
            food_locs.append((row, col))
        return food_locs

    def _generate_static_player_coords(self) -> List[Coord]:
        """Generate player start coords for static layout.

        Players always start around edge of field.
        """
        assert self._rows == self._cols

        assert len(self.possible_agents) <= math.ceil(self._rows / 2) * 4 - 4

        idxs = [
            i * 2 if i < self._rows / 4 else i * 2 + ((self._rows + 1) % 2)
            for i in range(math.ceil(self._rows / 2))
        ]
        idxs_reverse = list(reversed(idxs))

        # set it so locations are chosen from corners first then
        # alternating sides
        sides = [
            product([0], idxs[:-1]),
            product([self._rows - 1], idxs_reverse[:-1]),
            product(idxs_reverse[:-1], [0]),
            product(idxs[:-1], [self._rows - 1]),
        ]
        available_locations: List[Coord] = []
        for locs in zip(*sides):
            available_locations.extend(locs)

        return available_locations[: len(self.possible_agents)]

    def sample_initial_state(self) -> LBFState:
        if self._static_layout:
            players = self._spawn_players_static()
        else:
            players = self._spawn_players_generative()

        player_levels = sorted([player.level for player in players])
        max_level = sum(player_levels[:3])

        if self._static_layout:
            food = self._spawn_food_static(max_level)
        else:
            food = self._spawn_food_generative(
                self._max_food, max_level, [p.coord for p in players]
            )
        return LBFState(players, food, sum(f.level for f in food))

    def _spawn_players_static(self) -> Tuple[Player, ...]:
        assert self._player_locations is not None
        players = []
        for i in range(len(self.possible_agents)):
            (x, y) = self._player_locations[i]
            level = self.rng.randint(1, self._max_agent_level - 1)  # type: ignore
            players.append(Player(i, (x, y), level))
        return tuple(players)

    def _spawn_players_generative(self) -> Tuple[Player, ...]:
        players = []
        available_coords = list(product(range(self._cols), range(self._rows)))
        for i in range(len(self.possible_agents)):
            level = self.rng.randint(1, self._max_agent_level - 1)  # type: ignore
            coord = self.rng.choice(available_coords)
            available_coords.remove(coord)
            players.append(Player(i, coord, level))
        return tuple(players)

    def _spawn_food_static(self, max_level: int) -> Tuple[Food, ...]:
        """Spawn food in static layout.

        Number and location of food is the same for given pairing of
        (field size, max_food), only the levels of each food will change.
        """
        assert self._food_locations is not None
        assert max_level >= 1
        min_level = max_level if self._force_coop else 1
        food = []
        for x, y in self._food_locations:
            level = min_level
            if min_level != max_level:
                level = self.rng.randint(min_level, max_level)  # type: ignore
            food.append(Food((x, y), level))
        return tuple(food)

    def _spawn_food_generative(
        self,
        max_food: int,
        max_level: int,
        player_coords: List[Coord],
    ) -> Tuple[Food, ...]:
        assert max_level >= 1
        attempts = 0
        min_level = max_level if self._force_coop else 1
        unavailable_coords = set(player_coords)
        food: List[Food] = []
        while len(food) < max_food and attempts < 1000:
            attempts += 1
            x = self.rng.randint(1, self._cols - 2)  # type: ignore
            y = self.rng.randint(1, self._rows - 2)  # type: ignore

            if (x, y) in unavailable_coords:
                continue
            # add coord and adjacent coord to unavailable set so food are always
            # at least one cell apart
            for x_delta, y_delta in product([-1, 0, 1], repeat=2):
                unavailable_coords.add((x + x_delta, y + y_delta))
            level = min_level
            if min_level != max_level:
                level = self.rng.randint(min_level, max_level)  # type: ignore
            food.append(Food((x, y), level))
        return tuple(food)

    def sample_initial_obs(self, state: LBFState) -> Dict[M.AgentID, LBFObs]:
        return self._get_obs(state)

    def step(
        self, state: LBFState, actions: Dict[M.AgentID, LBFAction]
    ) -> M.JointTimestep[LBFState, LBFObs]:
        assert all(0 <= a < len(LBFAction) for a in actions.values())
        next_state, rewards = self._get_next_state(state, actions)
        obs = self._get_obs(next_state)
        all_done = len(next_state.food) == 0
        terminated = {i: all_done for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(
        self, state: LBFState, actions: Dict[M.AgentID, LBFAction]
    ) -> Tuple[LBFState, Dict[M.AgentID, float]]:
        next_food = {f.coord: f for f in state.food}

        # try move agents
        # if two or more agents try to move to the same location they all fail
        collisions = defaultdict(list)
        loading_players = set()
        for player in state.players:
            i = player.idx
            a_i = actions[str(i)]
            if a_i == LBFAction.NONE:
                next_coord = player.coord
            elif a_i == LBFAction.LOAD:
                next_coord = player.coord
                loading_players.add(player)
            else:
                next_coord = self.grid.get_next_coord(player.coord, ACTION_TO_DIR[a_i])
            if next_coord in next_food:
                next_coord = player.coord
            collisions[next_coord].append(player)

        # do movements for non colliding players
        next_players = list(state.players)
        for coord, player_list in collisions.items():
            # make sure no more than an player will arrive at location
            if len(player_list) > 1:
                continue
            player = player_list[0]
            next_players[player.idx] = Player(player.idx, coord, player.level)

        # process the loadings and calculate rewards as necessary
        rewards: Dict[M.AgentID, float] = {i: 0.0 for i in self.possible_agents}
        for player in loading_players:
            adj_coords = self.grid.get_neighbours(player.coord)
            for adj_coord in adj_coords:
                if adj_coord not in next_food:
                    continue
                adj_food = next_food[adj_coord]
                food_adj_coords = self.grid.get_neighbours(adj_coord)
                adj_players = [p for p in loading_players if p.coord in food_adj_coords]
                adj_player_level = sum([p.level for p in adj_players])

                if adj_player_level < adj_food.level:
                    # failed to load
                    for p in adj_players:
                        rewards[str(p.idx)] -= self._penalty
                else:
                    # food was loaded and each player scores points
                    for p in adj_players:
                        p_reward = float(p.level * adj_food.level)
                        if self._normalize_reward:
                            p_reward = p_reward / float(
                                adj_player_level * state.food_spawned
                            )
                        rewards[str(p.idx)] += p_reward
                    # and food is removed
                    next_food.pop(adj_coord)
        next_state = LBFState(
            tuple(next_players), tuple(next_food.values()), state.food_spawned
        )
        return next_state, rewards  # type: ignore

    def _get_obs(self, state: LBFState) -> Dict[M.AgentID, LBFObs]:
        obs: Dict[M.AgentID, LBFObs] = {}
        for i in self.possible_agents:
            player_obs, food_obs = self._get_local_obs(state, int(i))
            if self.observation_mode == "tuple":
                obs[i] = self._get_tuple_obs(player_obs, food_obs)
            elif self.observation_mode == "vector":
                obs[i] = self._get_vector_obs(int(i), player_obs, food_obs)
            else:
                obs[i] = self._get_grid_obs(
                    int(i), state.players[int(i)].coord, player_obs, food_obs
                )
        return obs

    def _get_local_obs(
        self, state: LBFState, agent_id: int
    ) -> Tuple[List[LBFEntityObs], List[LBFEntityObs]]:
        # player is always in center of observable area
        player_obs = []
        ego_player = state.players[agent_id]
        for p in state.players:
            if (
                abs(p.coord[0] - ego_player.coord[0]) <= self.sight
                and abs(p.coord[1] - ego_player.coord[1]) <= self.sight
            ):
                p_obs_coord = (
                    p.coord[0] - ego_player.coord[0] + self.sight,
                    p.coord[1] - ego_player.coord[1] + self.sight,
                )
                player_obs.append(
                    LBFEntityObs(p_obs_coord, p.level, is_self=(ego_player == p))
                )
            else:
                # player p not observed
                player_obs.append(LBFEntityObs((-1, -1), 0, is_self=False))

        food_obs = []
        for f in state.food:
            # note unobserved food are excluded since ordering can change for food obs
            # (unlike for players)
            if (
                abs(f.coord[0] - ego_player.coord[0]) <= self.sight
                and abs(f.coord[1] - ego_player.coord[1]) <= self.sight
            ):
                f_obs_coord = (
                    f.coord[0] - ego_player.coord[0] + self.sight,
                    f.coord[1] - ego_player.coord[1] + self.sight,
                )
                food_obs.append(LBFEntityObs(f_obs_coord, f.level, is_self=False))

        return player_obs, food_obs

    def _get_tuple_obs(
        self, player_obs: List[LBFEntityObs], food_obs: List[LBFEntityObs]
    ) -> LBFObs:
        obs: List[int] = []
        for o in player_obs:
            if o.is_self:
                obs.insert(0, o.level)
                obs.insert(0, o.coord[1])
                obs.insert(0, o.coord[0])
            else:
                obs.extend((*o.coord, o.level))
        for f_idx in range(self._max_food):
            if f_idx < len(food_obs):
                obs.extend((*food_obs[f_idx].coord, food_obs[f_idx].level))
            else:
                obs.extend((-1, -1, 0))
        return tuple(obs)

    def _get_vector_obs(
        self,
        agent_id: int,
        player_obs: List[LBFEntityObs],
        food_obs: List[LBFEntityObs],
    ) -> np.ndarray:
        # initialize obs array to (-1, -1, 0)
        obs = np.full(
            ((len(self.possible_agents) + self._max_food) * 3,), -1.0, dtype=np.float32
        )
        obs[2::3] = 0
        for i, o in enumerate(player_obs):
            if o.level == 0:
                # player unobserved
                continue
            p_idx = 0 if o.is_self else i * 3 if i > agent_id else (i + 1) * 3
            obs[p_idx : p_idx + 3] = (*o.coord, o.level)

        food_start_idx = len(self.possible_agents) * 3
        for i in range(self._max_food):
            if i < len(food_obs):
                f_idx = food_start_idx + (i * 3)
                obs[f_idx : f_idx + 3] = (*food_obs[i].coord, food_obs[i].level)

        return obs

    def _get_grid_obs(
        self,
        agent_id: int,
        agent_coord: Coord,
        player_obs: List[LBFEntityObs],
        food_obs: List[LBFEntityObs],
    ) -> np.ndarray:
        grid_shape_x, grid_shape_y = (2 * self.sight + 1, 2 * self.sight + 1)
        # agent, food, access layers
        grid_obs = np.zeros((3, grid_shape_x, grid_shape_y), dtype=np.float32)

        # add all in-bounds areas as accessible
        # cells containing players and food will be made inaccessible in next steps
        agent_x, agent_y = agent_coord
        for x, y in product(
            range(max(agent_x - self.sight, 0), min(self._cols, agent_x + self.sight)),
            range(max(agent_y - self.sight, 0), min(self._rows, agent_y + self.sight)),
        ):
            obs_x = x - agent_x + self.sight
            obs_y = y - agent_y + self.sight
            grid_obs[2, obs_x, obs_y] = 1.0

        for p in player_obs:
            if p.level > 0:
                grid_obs[0][p.coord] = p.level
                # agent coords are not accessible
                grid_obs[2][p.coord] = 0.0

        for f in food_obs:
            grid_obs[1][f.coord] = f.level
            # food coords are not accessible
            grid_obs[2][f.coord] = 0.0

        return grid_obs

    def parse_obs(
        self, obs: LBFObs
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        """Parse observation into (x, y, level) agent and food triplets.

        Agent obs are ordered so the observing agent is first, then the
        remaining observations are by agent order.

        On triplet of [-1, -1, 0] means no observation for the given agent or
        food.
        """
        if self.observation_mode == "tuple":
            assert isinstance(obs, tuple)
            return self.parse_tuple_obs(obs)

        assert isinstance(obs, np.ndarray)
        if self.observation_mode == "grid":
            return self.parse_grid_obs(obs)
        # vector obs
        return self.parse_vector_obs(obs)

    def parse_grid_obs(
        self, obs: np.ndarray
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        """Parse grid observation int (x, y, level) agent and food triplets.

        Agent obs are ordered so the observing agent is first, then the
        remaining observations are by agent order.

        On triplet of [-1, -1, 0] means no observation for the given agent or
        food.
        """
        raise NotImplementedError

    def parse_vector_obs(
        self, obs: np.ndarray
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        """Parse vector obs into (x, y, level) agent and food triplets.

        Agent obs are ordered so the observing agent is first, then the
        remaining observations are by agent order.

        On triplet of [-1, -1, 0] means no observation for the given agent or
        food.
        """
        assert obs.shape[0] == 3 * (len(self.possible_agents) + self._max_food)
        agent_obs = []
        food_obs = []
        for i in range(0, obs.shape[0], 3):
            triplet = (int(obs[i]), int(obs[i + 1]), int(obs[i + 2]))
            if i < len(self.possible_agents) * 3:
                agent_obs.append(triplet)
            else:
                food_obs.append(triplet)
        return agent_obs, food_obs

    def parse_tuple_obs(
        self, obs: Tuple[int, ...]
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        """Parse tuple obs into (x, y, level) agent and food triplets.

        Agent obs are ordered so the observing agent is first, then the
        remaining observations are by agent order.

        On triplet of [-1, -1, 0] means no observation for the given agent or
        food.
        """
        assert len(obs) == 3 * (len(self.possible_agents) + self._max_food)
        agent_obs = []
        food_obs = []
        for i in range(0, len(obs), 3):
            triplet = (int(obs[i]), int(obs[i + 1]), int(obs[i + 2]))
            if i < len(self.possible_agents) * 3:
                agent_obs.append(triplet)
            else:
                food_obs.append(triplet)
        return agent_obs, food_obs

    def get_obs_coords(self, origin: Coord) -> List[Coord]:
        """Get the list of coords observed from agent at origin."""
        obs_size = (2 * self.sight) + 1
        obs_coords: List[Coord] = []
        for col, row in product(range(obs_size), repeat=2):
            obs_grid_coord = self._map_obs_to_grid_coord((col, row), origin)
            if obs_grid_coord is not None:
                obs_coords.append(obs_grid_coord)
        return obs_coords

    def _map_obs_to_grid_coord(
        self, obs_coord: Coord, agent_coord: Coord
    ) -> Optional[Coord]:
        grid_col = agent_coord[0] + obs_coord[0] - self.sight
        grid_row = agent_coord[1] + obs_coord[1] - self.sight

        width, height = self.field_size
        if 0 <= grid_row < height and 0 <= grid_col < width:
            return (grid_col, grid_row)
        return None


def sorted_from_middle(lst):
    """Sorts list from middle out."""
    left = lst[len(lst) // 2 - 1 :: -1]
    right = lst[len(lst) // 2 :]
    output = [right.pop(0)] if len(lst) % 2 else []
    for t in zip(left, right):
        output += sorted(t)
    return output



"""The Unmanned Aerial Vehicle Grid World Environment."""
import random
from os import path
from typing import Dict, List, Optional, Set, Tuple, Union

from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.utils import seeding


UAVState = Tuple[Coord, Coord]
UAVAction = int

# UAV Obs = (uav coord, fug coord)
UAVUAVObs = Tuple[Coord, Coord]
# FUG Obs = house direction
UAVFUGObs = int
UAVObs = Union[UAVUAVObs, UAVFUGObs]

OBSNORTH = 0
OBSSOUTH = 1
OBSLEVEL = 2
OBSNONE = 3
DIR_OBS = [OBSNORTH, OBSSOUTH, OBSLEVEL, OBSNONE]
DIR_OBS_STR = ["N", "S", "L", "0"]


class UAVEnv(DefaultEnv[UAVState, UAVObs, UAVAction]):
    """The Unmanned Aerial Vehicle Grid World Environment.

    An adversarial 2D grid world problem involving two agents, a Unmanned
    Aerial Vehicle (UAV) and a fugitive. The UAV's goal is to capture the
    fugitive, while the fugitive's goal is to reach the safe house located at
    a known fixed location on the grid. The fugitive is considered caught if
    it is co-located with the UAV. The UAV observes it's own location and
    receives a noisy observation of the fugitive's location. The fugitive does
    not know it's location but it receives a noisy observation of its relative
    direction to the safe house when it is adjacent to the safe house.

    Posgible Agents
    ---------------
    - UAV = '0'
    - Fugitive = '1'

    State Space
    -----------
    Each state contains the `(x, y)` (x=column, y=row, with origin at the
    top-left square of the grid) of the UAV and fugitive agents. Specifically,
    a states is `((x_uav, y_uav), (x_fugitive, y_fugitive))`.

    Action Space
    ------------
    Each agent has 4 actions corresponding to moving in the 4 cardinal
    directions: `NORTH=0`, `EAST=1`, `SOUTH=2`, `WEST=3`.

    Observation Space
    -----------------
    The UAV observes it's `(x, y)` coordinates and receives a noisy observation
    of the fugitives `(x, y)` coordinates. The UAV observes the correct fugitive
    coordinates with probability `p=0.9`, and one of the adjacent locations to the true
    fugitive location with `p=1-0.9`.

    The fugitive can sense it's position with respect to the safe house, namely
    whether it is north of it (`OBSNORTH=0`), south of it (`OBSSOUTH=1`), or at the
    same level (`OBSLEVEL=3`). These observations are received with accuracy of `0.8`,
    and only when the fugitive is adjacent to it. If the fugitive is not
    adjacent to the safe house it receives no observation (`OBSNONE=4`).

    Rewards
    -------
    Both agents receive a penalty of `-0.04` for each step.
    If the fugitive reaches the safe house then the fugitive receives a reward of `1`,
    while the UAV receives a penalty of `-1`.
    If the fugitive is caught by the UAV, then the fugitive receives a penalty of `-1`,
    while the UAV receives a reward of `1`.

    Dynamics
    --------
    Actions are deterministic, so agents move into the next cell in the actions
    direction if it is not out-of-bounds. The fugitive's position is reset at random if
    it reaches the safe house or gets caught by the UAV.

    Starting State
    --------------
    Initially, the location of both agents is chosen at random, while the safe house
    is always located in the same location depending on grid layout.

    Episode End
    -----------
    Since the fugitive is reset each time it is caught. Episodes only end when the step
    limit is reached, as specified by  `max_episode_steps` when initializing the
    environment with `posggym.make` (default=`50`).

    Arguments
    ---------

    - `grid` - the grid of the environment. This can be an integer specifying
        the width and height of the grid, in which case an empty grid with the given
        dimensions and default position for the safe house will be used. Alternatively,
        it can be a `UAVGrid` instance. (default = `5`).

    Problem Sizes
    -------------

    For reference, the table below contains the size of state, and observation
    spaces for different grid sizes.

    | Grid size | States | Observations UAV | Observations Fugitive |
    |-----------|--------|------------------|-----------------------|
    | `3x3`     | 81     | 81               | 4                     |
    | `4x4`     | 256    | 256              | 4                     |
    | `5x5`     | 625    | 625              | 4                     |
    | `6x6`     | 1296   | 1296             | 4                     |

    Version History
    ---------------
    - `v0`: Initial version

    Reference
    ---------
    Panella, Alessandro, and Piotr Gmytrasiewicz. 2017. “Interactive POMDPs
    with Finite-State Models of Other Agents.” Autonomous Agents and
    Multi-Agent Systems 31 (4): 861–904.

    """

    metadata = {"render_modes": ["human", "ansi", "rgb_array"], "render_fps": 15}

    def __init__(
        self, grid: Union["UAVGrid", int] = 5, render_mode: Optional[str] = None
    ):
        super().__init__(UAVModel(grid), render_mode=render_mode)
        self.renderer = None
        self.uav_img = None
        self.fug_img = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return

        if self.render_mode == "ansi":
            return self._render_ansi()
        return self._render_gui()

    def _render_ansi(self):
        grid: UAVGrid = self.model.grid  # type: ignore
        grid_str = grid.get_ascii_repr(self._state[0], self._state[1])

        output = [
            f"Step: {self._step_num}",
            grid_str,
        ]
        if self._last_actions is not None:
            action_str = ", ".join(
                [str(Direction(a)) for a in self._last_actions.values()]
            )
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{self._last_rewards}>")
        return "\n".join(output) + "\n"

    def _render_gui(self):
        grid: UAVGrid = self.model.grid  # type: ignore

        import posggym.envs.grid_world.render as render_lib

        if self.renderer is None:
            self.renderer = render_lib.GWRenderer(
                self.render_mode,
                grid,
                render_fps=self.metadata["render_fps"],
                env_name="UAV",
                bg_color=(255, 255, 255),
                grid_line_color=(0, 0, 0),
            )

            # add house to static objects list
            img_path = path.join(path.dirname(__file__), "img", "house.png")
            img = render_lib.load_img_file(img_path, self.renderer.cell_size)
            house_img = render_lib.GWImage(
                grid.safe_house_coord, self.renderer.cell_size, img
            )
            self.renderer.static_objects.append(house_img)

        if self.uav_img is None:
            img_path = path.join(path.dirname(__file__), "img", "drone.png")
            img = render_lib.load_img_file(img_path, self.renderer.cell_size)
            self.uav_img = render_lib.GWImage((0, 0), self.renderer.cell_size, img)

        if self.fug_img is None:
            img_path = path.join(path.dirname(__file__), "img", "robber.png")
            img = render_lib.load_img_file(img_path, self.renderer.cell_size)
            self.fug_img = render_lib.GWImage((0, 0), self.renderer.cell_size, img)

        self.uav_img.coord = self._state[0]
        self.fug_img.coord = self._state[1]

        return self.renderer.render([self.uav_img, self.fug_img])

    def close(self):
        if self.renderer is not None:
            self.renderer.close()
            self.renderer = None


class UAVModel(M.POSGModel[UAVState, UAVObs, UAVAction]):
    """Unmanned Aerial Vehicle Problem Model."""

    NUM_AGENTS = 2

    UAV_ID = "0"
    FUG_ID = "1"

    R_ACTION = -0.04
    R_CAPTURE = 1.0  # UAV reward, fugitive = -R_CAPTURE
    R_SAFE = -1.0  # UAV reward, fugitive = -R_SAFE

    # Observatio Accuracy for each agent
    FUG_OBS_ACC = 0.8
    UAV_OBS_ACC = 0.9

    def __init__(self, grid: Union["UAVGrid", int]):
        if isinstance(grid, int):
            assert grid >= 3, "Grid size must be >= 3."
            # grid specified size of grid,
            grid = UAVGrid(grid, grid, None)
        self.grid = grid
        self._rng = random.Random(None)

        self.possible_agents = tuple(str(i) for i in range(self.NUM_AGENTS))
        self.state_space = spaces.Tuple(
            tuple(
                spaces.Tuple(
                    (
                        spaces.Discrete(self.grid.width),
                        spaces.Discrete(self.grid.height),
                    )
                )
                for _ in range(len(self.possible_agents))
            )
        )
        self.action_spaces = {
            i: spaces.Discrete(len(Direction)) for i in self.possible_agents
        }
        coord_space = spaces.Tuple(
            (spaces.Discrete(self.grid.width), spaces.Discrete(self.grid.height))
        )
        self.observation_spaces = {
            # UAV Obs = (uav coords, fug coords)
            self.UAV_ID: spaces.Tuple((coord_space, coord_space)),
            # FUG ubs = (safe house dir)
            self.FUG_ID: spaces.Discrete(len(DIR_OBS)),
        }
        self.is_symmetric = False

        # cache for sampling obs conditioned init state for fug
        self._cached_init_fug_obs: Optional[UAVFUGObs] = None
        self._valid_fug_coords_dist: Tuple[List[Coord], List[float]] = ([], [])

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[float, float]]:
        return {i: (self.R_SAFE, self.R_CAPTURE) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: UAVState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> UAVState:
        uav_coord = self.rng.choice(self.grid.init_uav_coords)
        fug_start_coords = list(self.grid.init_fug_coords)
        if uav_coord in fug_start_coords:
            fug_start_coords.remove(uav_coord)
        fug_coord = self.rng.choice(fug_start_coords)
        return uav_coord, fug_coord

    def sample_agent_initial_state(self, agent_id: M.AgentID, obs: UAVObs) -> UAVState:
        if agent_id == self.UAV_ID:
            assert isinstance(obs, tuple)
            return self._sample_uav_obs(obs)  # type: ignore

        assert isinstance(obs, int)
        house_adj_coords = self.grid.get_neighbours(
            self.grid.safe_house_coord, ignore_blocks=False
        )
        if len(house_adj_coords) != 4:
            # Doesn't work for 3x3 grid
            raise NotImplementedError(
                "Sampling observation conditioned initial state for the fugitive is "
                "only supported for maps where safe house can be reached from all "
                "directions."
            )
        return self._sample_fug(obs)

    def _sample_uav_initial_state(self, uav_obs: UAVUAVObs) -> UAVState:
        uav_coord, fug_coord = uav_obs
        if self.rng.random() < self.UAV_OBS_ACC:
            # UAV_OBS_ACC probability obs loc is correct
            true_fug_coord = fug_coord
        else:
            fug_start_coords = list(self.grid.init_fug_coords)
            if uav_coord in fug_start_coords:
                fug_start_coords.remove(uav_coord)
            true_fug_coord = self.rng.choice(fug_start_coords)
        return uav_coord, true_fug_coord

    def _sample_fug(self, fug_obs: UAVFUGObs) -> UAVState:
        if self._cached_init_fug_obs is None or self._cached_init_fug_obs != fug_obs:
            self._valid_fug_coords_dist = self._get_fug_coord_dist(fug_obs)

        valid_coords, cum_probs = self._valid_fug_coords_dist
        fug_coord = self.rng.choices(  # type: ignore
            valid_coords, cum_weights=cum_probs, k=1
        )[0]
        uav_start_coords = list(self.grid.init_uav_coords)
        if fug_coord in uav_start_coords:
            uav_start_coords.remove(fug_coord)
        uav_coord = self.rng.choice(uav_start_coords)
        return uav_coord, fug_coord

    def _get_fug_coord_dist(self, obs: UAVFUGObs) -> Tuple[List[Coord], List[float]]:
        house_adj_coords = self.grid.get_neighbours(
            self.grid.safe_house_coord, ignore_blocks=False
        )

        if obs == OBSNONE:
            valid_fug_coords = set(self.grid.init_fug_coords)
            valid_fug_coords.difference_update(house_adj_coords)
            p = 1.0 / len(valid_fug_coords)
            dist = []
            for i in range(len(valid_fug_coords)):
                if i == 0:
                    dist.append(p)
                else:
                    dist.append(dist[-1] + p)
            return list(valid_fug_coords), dist
        if obs == OBSNORTH:
            p = (1.0 - self.FUG_OBS_ACC) / (len(house_adj_coords) - 1)
            true_coords = [house_adj_coords[Direction.SOUTH]]
        elif obs == OBSSOUTH:
            p = (1.0 - self.FUG_OBS_ACC) / (len(house_adj_coords) - 1)
            true_coords = [house_adj_coords[Direction.NORTH]]
        elif obs == OBSLEVEL:
            p = (1.0 - self.FUG_OBS_ACC) / (len(house_adj_coords) - 2)
            true_coords = [
                house_adj_coords[Direction.EAST],
                house_adj_coords[Direction.WEST],
            ]

        dist = []
        num_adj, num_true = len(house_adj_coords), len(true_coords)
        for coord in house_adj_coords:
            if coord in true_coords:
                dist.append(self.FUG_OBS_ACC / num_true)
            else:
                dist.append((1.0 - self.FUG_OBS_ACC) / (num_adj - num_true))
        return house_adj_coords, dist

    def sample_initial_obs(self, state: UAVState) -> Dict[M.AgentID, UAVObs]:
        return self._sample_obs(state)

    def step(
        self, state: UAVState, actions: Dict[M.AgentID, UAVAction]
    ) -> M.JointTimestep[UAVState, UAVObs]:
        assert all(0 <= a_i < len(Direction) for a_i in actions.values())
        next_state = self._sample_next_state(state, actions)
        rewards = self._get_reward(next_state)

        uav_coord, fug_coord = next_state
        if fug_coord in (uav_coord, self.grid.safe_house_coord):
            # fug caught or fug safe
            fug_coord = self._sample_fug_coord(uav_coord)
            next_state = (uav_coord, fug_coord)

        obs = self._sample_obs(next_state)
        terminated = {i: False for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        all_done = False
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _sample_next_state(
        self, state: UAVState, actions: Dict[M.AgentID, UAVAction]
    ) -> UAVState:
        uav_a, fug_a = actions[self.UAV_ID], actions[self.FUG_ID]
        uav_coord, fug_coord = state
        uav_next_coord = self.grid.get_next_coord(uav_coord, Direction(uav_a))
        # fugitive resetting is handled in step function
        # to allow for
        if uav_next_coord == fug_coord:
            # UAV considered to capture Fugitive
            fug_next_coord = fug_coord
        else:
            fug_next_coord = self.grid.get_next_coord(fug_coord, Direction(fug_a))
        return uav_next_coord, fug_next_coord

    def _sample_fug_coord(self, uav_coord: Coord) -> Coord:
        fug_start_coords = list(self.grid.init_fug_coords)
        if uav_coord in fug_start_coords:
            fug_start_coords.remove(uav_coord)
        return self.rng.choice(fug_start_coords)

    def _sample_obs(self, state: UAVState) -> Dict[M.AgentID, UAVObs]:
        return {
            self.UAV_ID: self._sample_uav_obs(state),
            self.FUG_ID: self._sample_fug_obs(state),
        }

    def _sample_uav_obs(self, state: UAVState) -> UAVUAVObs:
        uav_coord, fug_coord = state
        if (
            fug_coord == self.grid.safe_house_coord
            or self.rng.random() < self.UAV_OBS_ACC
        ):
            fug_coord_obs = fug_coord
        else:
            adj_coords = self.grid.get_neighbours(fug_coord)
            if uav_coord in adj_coords:
                adj_coords.remove(uav_coord)
            fug_coord_obs = self.rng.choice(adj_coords)
        return uav_coord, fug_coord_obs

    def _sample_fug_obs(self, state: UAVState) -> UAVFUGObs:
        fug_coord = state[1]
        fug_adj_coords = self.grid.get_neighbours(
            fug_coord, ignore_blocks=True, include_out_of_bounds=True
        )

        safe_house_coord = self.grid.safe_house_coord
        if fug_adj_coords[0] == safe_house_coord:
            true_obs = OBSSOUTH
        elif fug_adj_coords[1] == safe_house_coord:
            true_obs = OBSNORTH
        elif safe_house_coord in fug_adj_coords[2:]:
            true_obs = OBSLEVEL
        else:
            true_obs = OBSNONE

        if true_obs == OBSNONE:
            return true_obs

        if self.rng.random() < self.FUG_OBS_ACC:
            return true_obs
        return self.rng.choice([OBSNORTH, OBSSOUTH, OBSLEVEL])

    def _get_reward(self, next_state: UAVState) -> Dict[M.AgentID, float]:
        uav_coord, fug_coord = next_state
        uav_reward, fug_reward = self.R_ACTION, self.R_ACTION
        if fug_coord == self.grid.safe_house_coord:
            uav_reward = self.R_SAFE
            fug_reward = -self.R_SAFE
        elif fug_coord == uav_coord:
            uav_reward = self.R_CAPTURE
            fug_reward = -self.R_CAPTURE
        return {
            self.UAV_ID: uav_reward,
            self.FUG_ID: fug_reward,
        }


class UAVGrid(Grid):
    """A grid for the MA UAV Problem."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Optional[Set[Coord]],
        safe_house_coord: Optional[Coord] = None,
        init_fug_coords: Optional[List[Coord]] = None,
        init_uav_coords: Optional[List[Coord]] = None,
    ):
        super().__init__(grid_width, grid_height, block_coords)
        if safe_house_coord is None:
            safe_house_coord = (grid_width // 2, grid_height // 4)
        self.safe_house_coord = safe_house_coord

        if init_fug_coords is None:
            init_fug_coords = self.all_coords
            init_fug_coords.remove(self.safe_house_coord)
        self.init_fug_coords = init_fug_coords

        if init_uav_coords is None:
            init_uav_coords = self.all_coords
            init_uav_coords.remove(self.safe_house_coord)
        self.init_uav_coords = init_uav_coords

        self.valid_coords = set(self.unblocked_coords)
        self.valid_coords.remove(self.safe_house_coord)

    def get_ascii_repr(
        self, fug_coord: Optional[Coord], uav_coord: Optional[Coord]
    ) -> str:
        """Get ascii repr of grid."""
        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord == self.safe_house_coord:
                    row_repr.append("S")
                elif coord in self.block_coords:
                    row_repr.append("#")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        if fug_coord is not None:
            grid_repr[fug_coord[0]][fug_coord[1]] = "F"
        if uav_coord is not None:
            if uav_coord == fug_coord:
                grid_repr[uav_coord[0]][uav_coord[1]] = "X"
            else:
                grid_repr[uav_coord[0]][uav_coord[1]] = "U"

        return str(self) + "\n" + "\n".join([" ".join(r) for r in grid_repr])



"""Utilities for storing and managing agent action-observation histories."""
from typing import Generic, List, Optional, Tuple

import posggym.model as M


class AgentHistory(Generic[M.ActType, M.ObsType]):
    """An Action-Observation history for a single agent in a POSG environment.

    A History is an ordered Tuple of (Action, Observation) tuples with one
    entry for each time step in the environment
    """

    def __init__(
        self, history: Tuple[Tuple[Optional[M.ActType], Optional[M.ObsType]], ...]
    ):
        self.history = history
        self.t = len(history) - 1

    def extend(self, action: M.ActType, obs: M.ObsType) -> "AgentHistory":
        """Extend the current history with given action, observation pair."""
        new_history = list(self.history)
        new_history.append((action, obs))
        return AgentHistory(tuple(new_history))

    def get_sub_history(self, horizon: int) -> "AgentHistory":
        """Get a subset of history up to given horizon."""
        assert 0 < horizon <= len(self.history), (
            "Cannot get sub history horizon must be 0 < horizon <= "
            f"len(history: 0 < {horizon} <= len(self.history) invalid"
        )
        if horizon == self.horizon:
            return self
        return AgentHistory(self.history[:horizon])

    def get_last_step(self) -> Tuple[Optional[M.ActType], Optional[M.ObsType]]:
        """Get the last step in the history."""
        return self.history[-1]

    @property
    def horizon(self) -> int:
        """Get the time horizon of this history.

        This is equal to the number of steps taken in the history
        """
        return len(self.history)

    @classmethod
    def get_init_history(cls, obs: Optional[M.ObsType] = None) -> "AgentHistory":
        """Get Initial history."""
        if obs is None:
            return cls(())
        return cls(((None, obs),))

    def __hash__(self):
        return hash(self.history)

    def __eq__(self, other):
        if not isinstance(other, AgentHistory):
            return False
        return self.history == other.history

    def __str__(self):
        h_str = []
        for a, o in self.history:
            h_str.append(f"({a}, {o})")
        h_str = ",".join(h_str)
        return f"h_{self.t}=<{h_str}>"

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, key):
        return self.history[key]

    def __iter__(self):
        return _AgentHistoryIterator(self)


class _AgentHistoryIterator:
    def __init__(self, history: AgentHistory):
        self.history = history
        self._idx = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._idx < len(self.history.history):
            self._idx += 1
            return self.history[self._idx - 1]
        raise StopIteration


class JointHistory:
    """A joint history for all agents in the environment."""

    def __init__(self, agent_histories: Tuple[AgentHistory, ...]):
        self.agent_histories = agent_histories
        self.num_agents = len(self.agent_histories)

    @classmethod
    def get_init_history(
        cls, num_agents: int, obs: Optional[Tuple[Optional[M.ObsType], ...]] = None
    ) -> "JointHistory":
        """Get Initial joint history."""
        if obs is None:
            return cls(
                tuple(AgentHistory.get_init_history() for _ in range(num_agents))
            )
        return cls(
            tuple(AgentHistory.get_init_history(obs[i]) for i in range(num_agents))
        )

    def get_agent_history(self, agent_id: int) -> AgentHistory:
        """Get the history of given agent."""
        return self.agent_histories[agent_id]

    def extend(
        self, action: Tuple[M.ActType, ...], obs: Tuple[M.ObsType, ...]
    ) -> "JointHistory":
        """Extend the current history with given action, observation pair."""
        new_agent_histories = []
        for i in range(self.num_agents):
            new_agent_histories.append(
                self.agent_histories[i].extend(action[i], obs[i])
            )
        return JointHistory(tuple(new_agent_histories))

    def get_sub_history(self, horizon: int) -> "JointHistory":
        """Get a subset of history up to given horizon."""
        sub_agent_histories = []
        for i in range(self.num_agents):
            sub_agent_histories.append(self.agent_histories[i].get_sub_history(horizon))
        return JointHistory(tuple(sub_agent_histories))

    def get_history_tm1(self) -> "JointHistory":
        """Get history at time t-1."""
        sub_agent_histories: List[AgentHistory] = []
        for i in range(self.num_agents):
            sub_agent_histories.append(AgentHistory(self.agent_histories[i][:-1]))
        return JointHistory(tuple(sub_agent_histories))

    def __hash__(self):
        return hash(self.agent_histories)

    def __eq__(self, other):
        if not isinstance(other, JointHistory):
            return False
        return self.agent_histories == other.agent_histories

    def __getitem__(self, key):
        actions = tuple(h[key][0] for h in self.agent_histories)
        obs = tuple(h[key][1] for h in self.agent_histories)
        return actions, obs

    def __str__(self):
        h_str = [f"{i} {h}" for i, h in enumerate(self.agent_histories)]
        h_str.insert(0, "<JointHistory:")
        h_str.append(">")
        return "\n".join(h_str)

    def __repr__(self):
        return self.__str__()



"""Wrapper for converting a posggym environment into rllib multi-agent environment."""
from typing import Optional, Set, Tuple

from gymnasium import spaces

import posggym


try:
    from ray.rllib.env.multi_agent_env import MultiAgentEnv
    from ray.rllib.utils.typing import AgentID, MultiAgentDict
except ImportError:
    raise posggym.error.DependencyNotInstalled(
        "The posggym.wrapper.rllib_multi_agent_env wrapper depends on the Ray RLlib "
        "library. run `pip install ray[rllib]>=2.3` or visit "
        "'https://docs.ray.io/en/latest/ray-overview/installation.html` for more "
        "details on installing rllib. "
    )


class RllibMultiAgentEnv(MultiAgentEnv):
    """An interface between a POSGGym env and a Rllib MultiAgentEnv.

    This involves:
    - treating agent IDs as strings (instead of ints)
    - handling case where an individual agent is done in environment before the episode
      is over

    Ref:
    https://github.com/ray-project/ray/blob/ray-2.3.0/rllib/env/multi_agent_env.py

    """

    def __init__(self, env: posggym.Env):
        self.env = env
        self._done_agents: Set[AgentID] = set()

        # must assign this first before calling super().__init__() so that
        # property functions are initialized before super().__init__() is
        # called
        self._agent_ids = set(self.env.possible_agents)
        # Do the action and observation spaces map from agent ids to spaces
        # for the individual agents?
        self._action_space_in_preferred_format = True
        self._obs_space_in_preferred_format = True

        super().__init__()

    @property
    def observation_space(self):
        """Get the environment observation space.

        This is a dictionary mapping agent IDs to the agent's observation space
        """
        return spaces.Dict(self.env.observation_spaces)

    @property
    def action_space(self):
        """Get the environment action space.

        This is a dictionary mapping agent IDs to the agent's action space
        """
        return spaces.Dict(self.env.action_spaces)

    def get_agent_ids(self) -> Set[AgentID]:
        """Return a set of agent ids in the environment."""
        return self._agent_ids

    def reset(  # type: ignore
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[MultiAgentDict, MultiAgentDict]:
        self._done_agents = set()
        return self.env.reset(seed=seed, options=options)

    def step(  # type: ignore
        self, action_dict: MultiAgentDict
    ) -> Tuple[
        MultiAgentDict, MultiAgentDict, MultiAgentDict, MultiAgentDict, MultiAgentDict
    ]:
        """Returns observations from ready agents.

        The returns are dicts mapping from agent_id strings to values. The
        number of agents in the env can vary over time.

        Returns:
        observations: new observations for each ready agent
        rewards: reward values for each ready agent. If the episode is just started, the
            value will be None.
        terminateds: terminated values for each ready agent. The special key "__all__"
            (required) is used to indicate env termination.
        truncateds: truncated values for each ready agent.
        infos: info values for each agent id (may be empty dicts).

        """
        obs, rewards, terminated, truncated, all_done, info = self.env.step(action_dict)

        # remove obs for already done agents, since pettingzoo expects no output for
        # agents that were terminated or truncated in a previous step
        for d in (obs, rewards, terminated, truncated, info):
            for i in self._done_agents:
                d.pop(i, None)

        # add any newly done agents to set of done agents so they are handled
        # appropriately next step
        for i, done in terminated.items():
            if done:
                self._done_agents.add(i)
        for i, done in truncated.items():
            if done:
                self._done_agents.add(i)

        # need to add RLlib specific special key to signify episode termination
        terminated["__all__"] = all_done
        truncated["__all__"] = all(truncated.values())

        return obs, rewards, terminated, truncated, info

    def render(self):
        output = self.env.render()
        if isinstance(output, dict):
            return output.get("env", None)
        return output

    def close(self):
        self.env.close()

    @property
    def unwrapped(self):
        return self



"""Model for the classic Multi-Agent Tiger problem."""
import sys
from itertools import product
from typing import Dict, List, Optional, Tuple

from gymnasium import spaces

import posggym.model as M
from posggym.core import DefaultEnv
from posggym.utils import seeding


MATState = int
TLEFT = 0
TRIGHT = 1
STATES = [TLEFT, TRIGHT]
STATE_STRS = ["TL", "TR"]

MATAction = int
OPENLEFT = 0
OPENRIGHT = 1
LISTEN = 2
ACTIONS = [OPENLEFT, OPENRIGHT, LISTEN]
ACTION_STR = ["OL", "OR", "L"]

MATObs = Tuple[int, int]
GROWLLEFT = 0
GROWLRIGHT = 1
CREAKLEFT = 0
CREAKRIGHT = 1
SILENCE = 2
OBS_PARTS = [(GROWLLEFT, GROWLRIGHT), (CREAKLEFT, CREAKRIGHT, SILENCE)]
OBS_SPACE = [
    (GROWLLEFT, CREAKLEFT),
    (GROWLLEFT, CREAKRIGHT),
    (GROWLLEFT, SILENCE),
    (GROWLRIGHT, CREAKLEFT),
    (GROWLRIGHT, CREAKRIGHT),
    (GROWLRIGHT, SILENCE),
]
OBS_STR = [("GL", "GR"), ("CL", "CR", "S")]

OTHER_AGENT_ID = {"0": "1", "1": "0"}


class MultiAgentTigerEnv(DefaultEnv):
    """The Multi-Agent Tiger Environment.

    This is a general-sum multi-agent version of the classic Tiger problem. It involves
    two agents that are in a corridor facing two doors: *left* and *right*. Behind one
    door lies a hungry tiger and behind the other lies treasure, but the agents do not
    know the position of either the tiger or the treasure. At each step each agent can
    choose to open one of the doors, or choose to listen, in which case it receives a
    noisy observation of if a door was opened and also the location of the tiger. If
    a door is opened the treasure and tiger are randomly reset.

    Possible Agents
    ---------------
    The environment supports two agents: '0' and '1'. Both agents are always active in
    the environment.

    State Space
    -----------
    The state is defined by which door the tiger is behind. `TLEFT=0` for tiger is
    behind the left door, and `TRIGHT=1` for the tiger is behind the right door.

    Action Space
    ------------
    Each agent can either open the left-hand door `OPENLEFT=0`, open the right-hand door
    `OPENRIGHT=1`, or listen for the presence of the tiger `LISTEN=2`.

    Observation Space
    -----------------
    Each agent observation consists of the tiger position observation and a door opening
    observation. The tiger position observation can be either: `GROWLEFT=0` for tiger
    left, `GROWLRIGHT=1` for tiger right. While the door opening observation can be
    either: `CREAKLEFT=0` for left door, `CREAKRIGHT=1` for right door, or `SILENCE=2`
    for silence.

    Combined each agent has 6 possible observations:

    1. `(GROWLEFT, CREAKLEFT)`
    2. `(GROWLEFT, CREAKRIGHT)`
    3. `(GROWLEFT, SILENCE)`
    4. `(GROWRIGHT, CREAKLEFT)`
    5. `(GROWRIGHT, CREAKRIGHT)`
    6. `(GROWRIGHT, SILENCE)`

    If an agent uses the `LISTEN=2` action they will perceive the correct current
    position of the tiger with probability `observation_prob` (default = `0.85`),
    independent of if the other agent opens a door or listens. Furthermore, they will
    perceive the correct door opening or not with probability `creak_observation_prob`
    (default = `0.9`).

    If an agent opens either door they will receive an observation uniformly at random
    from the set of all 6 possible observations.

    Rewards
    -------
    Each agent receives rewards independent of the other agent.

    An agent receives a reward of `+10` for opening the door without a tiger behind it,
    `-100` for opening the door with the tiger behind it, and `-1` for performing the
    listening action.

    Although the game is general-sum, and not zero-sum, agents influence each other by
    their effect on the state (i.e. resetting the tiger position).

    Dynamics
    --------
    The state is reset to `TLEFT` or `TRIGHT` with equal probability whenever either
    agent opens one of the doors. When both agents perform the `LISTEN` action - the
    state is unchanged.

    Starting State
    --------------
    The initial state is uniformly distributed between: `TLEFT` and `TRIGHT`.

    Episode End
    -----------
    By default episodes continue infinitely long. To set a step limit, specify
    `max_episode_steps` when initializing the environment with `posggym.make`.

    Arguments
    ---------

    - `observation_prob` - the probability of correctly observing the position of the
        tiger (default = `0.85`)
    - `creak_observation_prob` - the probability of correctly observing which door was
        opened by the other agent (default = `0.9`)

    Version History
    ---------------
    - `v0`: Initial version

    References
    ----------
    - Gmytrasiewicz, Piotr J., and Prashant Doshi. “A Framework for Sequential Planning
    in Multi-Agent Settings.” Journal of Artificial Intelligence Research 24 (2005).

    """

    metadata = {"render_modes": ["human", "ansi"], "render_fps": 4}

    def __init__(
        self,
        observation_prob: float = 0.85,
        creak_observation_prob: float = 0.9,
        render_mode: Optional[str] = None,
    ):
        super().__init__(
            MultiAgentTigerModel(observation_prob, creak_observation_prob),
            render_mode=render_mode,
        )

    def render(self):
        assert self._last_obs is not None

        state_str = STATE_STRS[self._state]
        obs_str = ", ".join(
            [str((OBS_STR[0][o[0]], OBS_STR[1][o[1]])) for o in self._last_obs.values()]
        )
        output = [
            f"Step: {self._step_num}",
            f"State: <{state_str}>",
            f"Obs: <{obs_str}>",
        ]
        if self._last_actions is not None:
            action_str = ", ".join([ACTION_STR[a] for a in self._last_actions.values()])
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{tuple(self._last_rewards.values())}>")

        output_str = "\n".join(output) + "\n"

        if self.render_mode == "human":
            sys.stdout.write(output_str)
        else:
            # ansi mode
            return output_str


class MultiAgentTigerModel(M.POSGFullModel[MATState, MATObs, MATAction]):
    """Multi-Agent Tiger problem Model."""

    NUM_AGENTS = 2

    OPEN_GOOD_R = 10.0
    OPEN_BAD_R = -100.0
    LISTEN_R = -1.0

    def __init__(
        self, observation_prob: float = 0.85, creak_observation_prob: float = 0.9
    ):
        assert 0 <= observation_prob <= 1.0
        assert 0 <= creak_observation_prob <= 1.0
        self._obs_prob = observation_prob
        self._creak_obs_prob = creak_observation_prob

        self.possible_agents = tuple(str(i) for i in range(self.NUM_AGENTS))
        self.state_space = spaces.Discrete(len(STATES))
        self.action_spaces = {
            i: spaces.Discrete(len(ACTIONS)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: spaces.Tuple(
                (spaces.Discrete(len(OBS_PARTS[0])), spaces.Discrete(len(OBS_PARTS[1])))
            )
            for i in self.possible_agents
        }
        self.is_symmetric = True

        # Spaces used internally
        self._state_space = STATES
        self._action_spaces = tuple(
            [*ACTIONS] for _ in range(len(self.possible_agents))
        )
        self._obs_spaces = tuple(OBS_SPACE for _ in range(len(self.possible_agents)))

        self._trans_map = self._construct_trans_func()
        self._rew_map = self._construct_rew_func()
        self._obs_map = self._construct_obs_func()

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[float, float]]:
        return {i: (self.OPEN_BAD_R, self.OPEN_BAD_R) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: MATState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> MATState:
        return self.rng.choice(STATES)

    def sample_initial_obs(self, state: MATState) -> Dict[M.AgentID, MATObs]:
        return {i: OBS_SPACE[2] for i in self.possible_agents}

    def step(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> M.JointTimestep[MATState, MATObs]:
        assert all(a_i in ACTIONS for a_i in actions.values())
        next_state = self._sample_next_state(state, actions)
        obs = self._sample_obs(state, actions)
        rewards = self._get_reward(state, actions)
        terminated = {i: False for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        all_done = False
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _sample_next_state(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> MATState:
        next_state = state
        if any(a != LISTEN for a in actions.values()):
            next_state = self.rng.choice(STATES)
        return next_state

    def _sample_obs(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> Dict[M.AgentID, MATObs]:
        obs: Dict[M.AgentID, MATObs] = {}
        for i, a in actions.items():
            if a != LISTEN:
                obs[i] = self.rng.choice(OBS_SPACE)
                continue

            other_agent_id = OTHER_AGENT_ID[i]
            other_agent_action = actions[other_agent_id]

            tiger_obs = self._sample_tiger_obs(state)
            creak_obs = self._sample_creak_obs(other_agent_action)
            obs[i] = (tiger_obs, creak_obs)
        return obs

    def _sample_tiger_obs(self, state: MATState) -> int:
        if self.rng.random() < self._obs_prob:
            # correct obs
            return GROWLLEFT if state == TLEFT else GROWLRIGHT
        return GROWLRIGHT if state == TLEFT else GROWLLEFT

    def _sample_creak_obs(self, a_j: MATAction) -> int:
        if self.rng.random() < self._creak_obs_prob:
            # correct obs
            if a_j == LISTEN:
                return SILENCE
            return CREAKLEFT if a_j == OPENLEFT else CREAKRIGHT

        if a_j == LISTEN:
            return self.rng.choice([CREAKLEFT, CREAKRIGHT])
        if a_j == OPENLEFT:
            return self.rng.choice([CREAKRIGHT, SILENCE])
        return self.rng.choice([CREAKRIGHT, SILENCE])

    def _get_reward(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> Dict[M.AgentID, float]:
        rewards: Dict[M.AgentID, float] = {}
        for i, a in actions.items():
            if a == LISTEN:
                rewards[i] = self.LISTEN_R
            elif a == state:
                rewards[i] = self.OPEN_BAD_R
            else:
                rewards[i] = self.OPEN_GOOD_R
        return rewards

    def get_initial_belief(self) -> Dict[MATState, float]:
        b_map: Dict[MATState, float] = {}
        for s in STATES:
            s_prob = 1.0 / len(STATES)
            b_map[s] = s_prob
        return b_map

    def transition_fn(
        self, state: MATState, actions: Dict[M.AgentID, MATAction], next_state: MATState
    ) -> float:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._trans_map[(state, action_tuple, next_state)]

    def _construct_trans_func(self) -> Dict:
        trans_map = {}
        uniform_prob = 1.0 / len(STATES)
        for s, a, s_next in product(
            self._state_space, product(*self._action_spaces), self._state_space
        ):
            trans_map[(s, a, s_next)] = (
                uniform_prob if any(a_i != LISTEN for a_i in a) else float(s_next == s)
            )
        return trans_map

    def observation_fn(
        self,
        obs: Dict[M.AgentID, MATObs],
        next_state: MATState,
        actions: Dict[M.AgentID, MATAction],
    ) -> float:
        obs_tuple = tuple(obs[i] for i in self.possible_agents)
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._obs_map[(next_state, action_tuple, obs_tuple)]

    def _construct_obs_func(self) -> Dict:
        obs_func = {}
        uniform_o_prob = 1.0 / len(OBS_SPACE)
        for s_next, a, o in product(
            self._state_space, product(*self._action_spaces), product(*self._obs_spaces)
        ):
            o_prob = 1.0
            a_i, a_j = a
            o_i, o_j = o
            correct_pos = GROWLLEFT if s_next == TLEFT else GROWLRIGHT
            if (a_i, a_j) == (LISTEN, LISTEN):
                for o_k in (o_i, o_j):
                    if o_k[0] == correct_pos:
                        o_prob *= self._obs_prob
                    else:
                        o_prob *= 1.0 - self._obs_prob

                    if o_k[1] == SILENCE:
                        o_prob *= self._creak_obs_prob
                    else:
                        o_prob *= (1.0 - self._creak_obs_prob) / 2

            elif LISTEN in (a_i, a_j):
                if a_i == LISTEN:
                    a_k, o_k = a_j, o_i
                else:
                    a_k, o_k = a_i, o_j

                o_prob *= uniform_o_prob
                correct_creak = CREAKLEFT if a_k == OPENLEFT else CREAKRIGHT

                if o_k[0] == correct_pos:
                    o_prob *= self._obs_prob
                else:
                    o_prob *= 1.0 - self._obs_prob

                if o_k[1] == correct_creak:
                    o_prob *= self._creak_obs_prob
                else:
                    o_prob *= (1.0 - self._creak_obs_prob) / 2

            else:
                o_prob *= uniform_o_prob * uniform_o_prob

            obs_func[(s_next, a, o)] = o_prob

        return obs_func

    def reward_fn(
        self, state: MATState, actions: Dict[M.AgentID, MATAction]
    ) -> Dict[M.AgentID, float]:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._rew_map[(state, action_tuple)]

    def _construct_rew_func(self) -> Dict:
        rew_map = {}
        joint_actions_space = product(*self._action_spaces)
        for s, a in product(self._state_space, joint_actions_space):
            rew_map[(s, a)] = self._get_reward(
                s, {str(i): a_i for i, a_i in enumerate(a)}
            )
        return rew_map



"""A set of functions for checking a model's details.

Based on module in the Gymnasium repository:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/utils/env_checker.py

Which in turn is based on work from:

- Stable Baselines3 repository hosted on GitHub
  (https://github.com/DLR-RM/stable-baselines3/). Original Author: Antonin Raffin.
- warnings/assertions from the PettingZoo repository hosted on GitHub
  (https://github.com/PettingZoo-Team/PettingZoo). Original Author: J K Terry/.

These projects are covered by the MIT License.

"""
import inspect
from copy import deepcopy
from typing import Optional

import posggym.model as M
from posggym import logger
from posggym.utils.passive_env_checker import (
    check_rng_equality,
    data_equivalence,
    check_agent_action_spaces,
    check_agent_obs,
    check_agent_observation_spaces,
    check_agent_space_limits,
    check_state,
    check_state_space,
    model_step_passive_checker,
)


def check_initial_state_type(model: M.POSGModel):
    """Checks that :meth:`sample_initial_state` correctly returns a valid state.

    Arguments
    ---------
    model: The model to check

    Returns
    -------
    state: sampled initial state

    Raises
    ------
    AssertionError depending on spec violation

    """
    state = model.sample_initial_state()
    check_state(state, model)
    return state


def check_initial_obs_type(model: M.POSGModel, state: Optional[M.StateType] = None):
    """Checks that :meth:`sample_initial_obs` works correctly.

    Assumes ``model.sample_initial_state()`` works as expected.

    Arguments
    ---------
    model: the model to check
    state: the state to use for check, default is None in which case a new state is
           sampled from the model.

    Raises
    ------
    AssertionError depending on spec violation

    """
    if state is None:
        state = model.sample_initial_state()

    try:
        obs = model.sample_initial_obs(state)
        check_agent_obs(obs, model.observation_spaces, "sample_initial_obs")
    except NotImplementedError:
        raise AssertionError(
            "Model requires the ``sample_initial_obs`` method to be implemented."
        )


def check_initial_sampling_seed(model: M.POSGModel):
    """Check that model seeding works correctly for initial conditions.

    Arguments
    ---------
    model: The environment model to check

    Raises
    ------
    AssertionError: The model random seeding doesn't work as expected.

    """
    signature = inspect.signature(model.seed)
    assert "seed" in signature.parameters or (
        "kwargs" in signature.parameters
        and signature.parameters["kwargs"].kind is inspect.Parameter.VAR_KEYWORD
    ), "The `seed` method does not provide a `seed` or `**kwargs` keyword argument."
    try:
        model.seed(seed=123)
        state_1 = model.sample_initial_state()
        check_state(state_1, model)
        obs_1 = model.sample_initial_obs(state_1)
        check_agent_obs(obs_1, model.observation_spaces, "sample_initial_obs")

        assert model._rng is not None, (  # pyright: ignore [reportPrivateUsage]
            "Expects the random number generator to have been generated given a "
            "seed was passed to the `seed` method. Mostly likely the model `seed` "
            "method does not call `super().seed(seed=seed)` or the `rng` property has "
            "not been implemented."
        )
        seed_123_rng = deepcopy(model._rng)  # pyright: ignore [reportPrivateUsage]

        model.seed(seed=123)
        state_2 = model.sample_initial_state()
        check_state(state_2, model)
        obs_2 = model.sample_initial_obs(state_2)
        check_agent_obs(obs_2, model.observation_spaces, "sample_initial_obs")

        if model.spec is not None and model.spec.nondeterministic is False:
            assert data_equivalence(state_1, state_2), (
                "Using `model.seed(seed=123)` is non-deterministic as the "
                "initial states are not equivalent."
            )
            assert data_equivalence(obs_1, obs_2), (
                "Using `model.seed(seed=123)` is non-deterministic as the "
                "initial observations are not equivalent."
            )

        check_rng_equality(
            model._rng,  # pyright: ignore [reportPrivateUsage]
            seed_123_rng,
            prefix=(
                "Mostly likely the model seed function does not call "
                "`super().seed(seed=seed)` as the random generators are not same "
                "when the same seeds are passed to `model.seed`. "
            ),
        )

        model.seed(seed=456)
        state_3 = model.sample_initial_state()
        check_state(state_3, model)
        obs_3 = model.sample_initial_obs(state_3)
        check_agent_obs(obs_3, model.observation_spaces, "sample_initial_obs")

        try:
            check_rng_equality(
                model._rng,  # pyright: ignore [reportPrivateUsage]
                seed_123_rng,
                prefix="",
            )
        except AssertionError:
            pass
        else:
            raise AssertionError(
                "Mostly likely the environment reset function does not call "
                "`super().reset(seed=seed)` as the random number generators are not"
                " different when different seeds are passed to `env.reset`."
            )

    except TypeError as e:
        raise AssertionError(
            "The environment cannot be reset with a random seed, even though "
            "`seed` or `kwargs` appear in the signature. This should never happen, "
            f"please report this issue. The error was: {e}"
        ) from e

    seed_param = signature.parameters.get("seed")
    # Check the default value is None
    if seed_param is not None and seed_param.default is not None:
        logger.warn(
            "The default seed argument in `seed` method should be `None`, otherwise "
            "the model will by default always be deterministic. "
            f"Actual default: {seed_param.default}"
        )


def check_model(model: M.POSGModel):
    """Check that an environment model follows posggym API.

    This is an invasive function that calls the models step.

    This is particularly useful when using a custom environment.

    TODO Update links

    Please take a look at https://gymnasium.farama.org/content/environment_creation/
    for more information about the API.

    Arguments
    ---------
    model: The posggym environment model that will be checked

    """
    more_info_msg = "See COMING SOON for more info."

    assert isinstance(
        model, M.POSGModel
    ), f"The model must inherit from the posggym.POSGModel class. {more_info_msg}"

    # ============= Check agents =============
    assert hasattr(
        model, "possible_agents"
    ), f"The model must specify possible agents tuple. {more_info_msg}"
    assert isinstance(model.possible_agents, tuple), (
        "Model `possible_agents` should be tuple (so it's immutable). "
        f"Actual type: {type(model.possible_agents)}."
    )
    assert all(isinstance(i, M.AgentID) for i in model.possible_agents), (
        f"Model agent IDs in `possible_agents` must be have type {M.AgentID}. "
        f"Actual types: {[type(i) for i in model.possible_agents]}."
    )

    # ============= Check basic model properties =============
    assert hasattr(
        model, "is_symmetric"
    ), f"The model must specify whether it's symmetric or not. {more_info_msg}"
    assert isinstance(
        model.is_symmetric, bool
    ), f"Model `is_symmetric` should be bool. Actual type: {type(model.is_symmetric)}."
    assert all(i in model.reward_ranges for i in model.possible_agents), (
        "Model `reward_ranges` should be defined for all agents in `possible_agents`."
        "Missing: "
        f"{[i for i in model.possible_agents if i not in model.reward_ranges]}."
    )

    # ============= Check the spaces (action, observation and state) =============
    assert hasattr(
        model, "action_spaces"
    ), f"The model must specify agent action spaces. {more_info_msg}"
    check_agent_action_spaces(model.action_spaces)
    check_agent_space_limits(model.action_spaces, "action")

    assert hasattr(
        model, "observation_spaces"
    ), f"The environment must specify agent observation spaces. {more_info_msg}"
    check_agent_observation_spaces(model.observation_spaces)
    check_agent_space_limits(model.observation_spaces, "observation")

    if hasattr(model, "state_space") and model.state_space is not None:
        check_state_space(model.state_space)

    # ============= Check sampling initial states and observations =============
    check_initial_state_type(model)
    check_initial_obs_type(model)
    check_initial_sampling_seed(model)

    # ============ Check the returned values ===============
    state = model.sample_initial_state()
    model_step_passive_checker(
        model,
        state,
        {i: model.action_spaces[i].sample() for i in model.get_agents(state)},
    )



"""The Two-Paths Grid World Environment."""
import itertools
from os import path
from typing import Dict, List, Optional, Set, Tuple, Union

from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.utils import seeding


TPState = Tuple[Coord, Coord]
TPAction = int
# Obs = adj_obs
TPObs = Tuple[int, int, int, int]

# Cell obs
OPPONENT = 0
WALL = 1
EMPTY = 2

CELL_OBS = [OPPONENT, WALL, EMPTY]
CELL_OBS_STR = ["X", "#", "0"]


class TwoPathsEnv(DefaultEnv[TPState, TPObs, TPAction]):
    """The Two-Paths Grid World Environment.

    An adversarial 2D grid world problem involving two agents, a runner and
    a chaser. The runner's goal is to reach one of two goal location, with each
    goal located at the end of a separate path. The lengths of the two paths
    have different lengths. The goal of the chaser is to intercept the runner
    before it reaches a goal. The runner is considered caught if it is observed
    by the chaser, or occupies the same location. However, the chaser is only
    able to effectively cover one of the two goal locations.

    This environment requires each agent to reason about the which path the
    other agent will choose. It offers an ideal testbed for planning under
    finite-nested reasoning assumptions since it is possible to map reasoning
    level to the expected path choice.

    The two agents start at opposite ends of the maps.

    Possible Agents
    ---------------
    - Runner = '0'
    - Chaser = '1'

    State Space
    -----------
    Each state contains the `(x, y)` (x=column, y=row, with origin at the
    top-left square of the grid) of the runner and chaser agent. Specifically,
    a state is `((x_runner, y_runner), (x_chaser, y_chaser))`

    Action Space
    ------------
    Each agent has 4 actions corresponding to moving in the 4 cardinal
    directions: `NORTH=0`, `EAST=1`, `SOUTH=2`, `WEST=3`.

    Observation Space
    -----------------
    Each agent observes the adjacent cells in the four cardinal directions and
    whether they are one of three things: `OPPONENT=0`, `WALL=1`, `EMPTY=2`.

    Each observation is represented as a tuple:
    `(cell_north, cell_south, cell_east, cell_west)`

    Rewards
    -------
    Both agents receive a penalty of `-0.01` for each step. If the runner reaches the
    goal then the runner receives a reward of `1.0`, while the chaser receives a penalty
    of `-1.0`. If the runner is observed by the chaser, then the runner receives a
    penalty of `-1.0`, while the chaser receives a reward of `1.0`.

    The rewards make the environment adversarial, but not strictly zero-sum, due to the
    small penalty each step.

    Dynamics
    --------
    By default actions are deterministic and will lead to the agent moving one cell in
    the action's direction, given the cell adjacent cell in that direction is empty and
    not out-of-bounds.

    The environment can also be run in stochastic mode by changing the `action_probs`
    parameter at initialization. This controls the probability the agent will move in
    the desired direction each step, otherwise moving randomly in one of the other 3
    other possible directions.

    Starting State
    --------------
    Both runner and chaser agents start at the same location on opposite ends of the
    grid each episode, and the goal locations and grid layout are also the same. The
    specific starting state and configuration depends on the grid layout used.

    Episodes End
    ------------
    Episode ends when either the runner is caught, or reaches a goal. By default a
    `max_episode_steps` limit of `20` is also set.

    Arguments
    ---------

    - `grid_size` - the grid size to use. This can either `3`, `4`, or `7`, each size
        `n` create a TwoPaths Env with a `n`-by-`n` grid layout (default = `7`).
    - `action_probs` - the action success probability for each agent. This can be a
        single float (same value for both runner and chaser agents) or a tuple with
        separate values for the runner and chaser (default = `1.0`).

    Reference
    ---------
    Schwartz, Jonathon, Ruijia Zhou, and Hanna Kurniawati. "Online Planning for
    Interactive-POMDPs using Nested Monte Carlo Tree Search." In 2022 IEEE/RSJ
    International Conference on Intelligent Robots and Systems (IROS), pp. 8770-8777.
    IEEE, 2022.

    """

    metadata = {
        "render_modes": ["human", "ansi", "rgb_array", "rgb_array_dict"],
        "render_fps": 15,
    }

    def __init__(
        self,
        grid_size: int = 7,
        action_probs: Union[float, Tuple[float, float]] = 1.0,
        render_mode: Optional[str] = None,
    ):
        super().__init__(
            TwoPathsModel(grid_size, action_probs),
            render_mode=render_mode,
        )
        self.renderer = None
        self.chaser_img = None
        self.runner_img = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return

        if self.render_mode == "ansi":
            return self._render_ansi()
        return self._render_img()

    def _render_ansi(self):
        grid = self.model.grid  # type: ignore
        grid_str = grid.get_ascii_repr(self._state[0], self._state[1])

        output = [
            f"Step: {self._step_num}",
            grid_str,
        ]
        if self._last_actions is not None:
            action_str = ", ".join(
                [str(Direction(a)) for a in self._last_actions.values()]
            )
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{self._last_rewards}>")

        return "\n".join(output) + "\n"

    def _render_img(self):
        grid: Grid = self.model.grid  # type: ignore

        import posggym.envs.grid_world.render as render_lib

        if self.renderer is None:
            self.renderer = render_lib.GWRenderer(
                self.render_mode,
                grid,
                render_fps=self.metadata["render_fps"],
                env_name="Two Paths",
            )

            goal_imgs = [
                render_lib.GWRectangle(
                    coord, self.renderer.cell_size, render_lib.get_color("green")
                )
                for coord in grid.goal_coords
            ]
            self.renderer.static_objects.extend(goal_imgs)

        if self.runner_img is None:
            img_path = path.join(path.dirname(__file__), "img", "robot.png")
            img = render_lib.load_img_file(img_path, self.renderer.cell_size)
            self.runner_img = render_lib.GWImage((0, 0), self.renderer.cell_size, img)

        if self.chaser_img is None:
            img_path = path.join(path.dirname(__file__), "img", "robot.png")
            img = render_lib.load_img_file(img_path, self.renderer.cell_size)
            self.chaser_img = render_lib.GWImage((0, 0), self.renderer.cell_size, img)

        self.runner_img.coord = self._state[0]
        self.chaser_img.coord = self._state[1]
        render_objects = [self.runner_img, self.chaser_img]

        observed_coords = grid.get_neighbours(self._state[0])
        observed_coords.extend(grid.get_neighbours(self._state[1]))

        if self.render_mode in ("human", "rgb_array"):
            return self.renderer.render(render_objects, observed_coords)

        return self.renderer.render_agents(
            render_objects,
            {
                "0": (self._state[0], Direction.NORTH),
                "1": (self._state[1], Direction.NORTH),
            },
            agent_obs_dims=1,
            observed_coords=observed_coords,
            # mask corners of obs square
            agent_obs_mask=[(0, 0), (0, 2), (2, 0), (2, 2)],
        )

    def close(self) -> None:
        if self.renderer is not None:
            self.renderer.close()
            self.renderer = None


class TwoPathsModel(M.POSGModel[TPState, TPObs, TPAction]):
    """Two-Paths Problem Model.

    Parameters
    ----------
    grid_name : str
        determines the environment map (see posggym.envs.two_paths.grid for
        available grids)
    action_probs : float or (float, float)
        the action success probability for each agent. By default the
        environment is deterministic (action_probs=1.0).

    """

    NUM_AGENTS = 2

    RUNNER_IDX = 0
    CHASER_IDX = 1

    R_ACTION = -0.01
    R_CAPTURE = -1.0  # Runner reward, Chaser = -R_CAPTURE
    R_SAFE = 1.0  # Runner reward, Chaser = -R_SAFE

    def __init__(
        self,
        grid_size: int,
        action_probs: Union[float, Tuple[float, float]] = 1.0,
    ):
        assert grid_size in SUPPORTED_GRIDS, (
            f"Unsupported grid_size of `{grid_size}`, must be one of: "
            f"{SUPPORTED_GRIDS.keys()}."
        )
        self.grid = SUPPORTED_GRIDS[grid_size]()

        if isinstance(action_probs, float):
            action_probs = (action_probs, action_probs)
        self._action_probs = action_probs

        self.possible_agents = tuple(str(i) for i in range(self.NUM_AGENTS))
        self.state_space = spaces.Tuple(
            tuple(
                spaces.Tuple(
                    (
                        spaces.Discrete(self.grid.width),
                        spaces.Discrete(self.grid.height),
                    )
                )
                for _ in range(len(self.possible_agents))
            )
        )
        self.action_spaces = {
            i: spaces.Discrete(len(Direction)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: spaces.Tuple(
                (
                    spaces.Discrete(len(CELL_OBS)),
                    spaces.Discrete(len(CELL_OBS)),
                    spaces.Discrete(len(CELL_OBS)),
                    spaces.Discrete(len(CELL_OBS)),
                )
            )
            for i in self.possible_agents
        }
        self.is_symmetric = False

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[float, float]]:
        return {i: (self.R_CAPTURE, self.R_SAFE) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: TPState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> TPState:
        return (self.grid.init_runner_coord, self.grid.init_chaser_coord)

    def sample_agent_initial_state(self, agent_id: M.AgentID, obs: TPObs) -> TPState:
        return self.sample_initial_state()

    def get_initial_belief_dist(self) -> Dict[TPState, float]:
        s_0 = (self.grid.init_runner_coord, self.grid.init_chaser_coord)
        return {
            s: float(s == s_0)  # type: ignore
            for s in itertools.product(self.grid.all_coords, repeat=2)
        }

    def sample_initial_obs(self, state: TPState) -> Dict[M.AgentID, TPObs]:
        return self._get_obs(state)

    def step(
        self, state: TPState, actions: Dict[M.AgentID, TPAction]
    ) -> M.JointTimestep[TPState, TPObs]:
        assert all(0 <= a_i < len(Direction) for a_i in actions.values())
        next_state = self._get_next_state(state, actions)
        rewards = self._get_rewards(next_state)
        terminal = self._state_is_terminal(next_state)

        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        if terminal:
            for i, outcome in self._get_outcome(next_state).items():
                info[i]["outcome"] = outcome

        obs = self._get_obs(next_state)
        truncated = {i: False for i in self.possible_agents}
        terminated = {i: terminal for i in self.possible_agents}

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, terminal, info
        )

    def _get_next_state(
        self, state: TPState, actions: Dict[M.AgentID, TPAction]
    ) -> TPState:
        runner_coord = state[self.RUNNER_IDX]
        chaser_coord = state[self.CHASER_IDX]
        runner_a = actions[str(self.RUNNER_IDX)]
        chaser_a = actions[str(self.CHASER_IDX)]

        if self.rng.random() > self._action_probs[self.CHASER_IDX]:
            other_as = [a for a in Direction if a != chaser_a]
            chaser_a = self.rng.choice(other_as)
        chaser_next_coord = self.grid.get_next_coord(
            chaser_coord, Direction(chaser_a), ignore_blocks=False
        )

        if chaser_next_coord == runner_coord:
            # Runner considered to capture Fugitive
            runner_next_coord = runner_coord
        else:
            if self.rng.random() > self._action_probs[self.RUNNER_IDX]:
                other_as = [a for a in Direction if a != runner_a]
                runner_a = self.rng.choice(other_as)
            runner_next_coord = self.grid.get_next_coord(
                runner_coord, Direction(runner_a), ignore_blocks=False
            )

        return (runner_next_coord, chaser_next_coord)

    def _get_obs(self, state: TPState) -> Dict[M.AgentID, TPObs]:
        runner_coord = state[self.RUNNER_IDX]
        chaser_coord = state[self.CHASER_IDX]
        return {
            str(self.RUNNER_IDX): self._get_adj_obs(runner_coord, chaser_coord),
            str(self.CHASER_IDX): self._get_adj_obs(chaser_coord, runner_coord),
        }

    def _get_adj_obs(
        self, coord: Coord, opponent_coord: Coord
    ) -> Tuple[int, int, int, int]:
        adj_obs = []
        for d in Direction:
            next_coord = self.grid.get_next_coord(coord, d, False)
            if next_coord == opponent_coord:
                adj_obs.append(OPPONENT)
            elif next_coord == coord:
                adj_obs.append(WALL)
            else:
                adj_obs.append(EMPTY)
        return tuple(adj_obs)  # type: ignore

    def _get_rewards(self, state: TPState) -> Dict[M.AgentID, float]:
        runner_coord = state[self.RUNNER_IDX]
        chaser_coord = state[self.CHASER_IDX]
        r_runner, r_chaser = (self.R_ACTION, self.R_ACTION)
        if runner_coord in self.grid.goal_coords:
            r_runner, r_chaser = (self.R_SAFE, -self.R_SAFE)
        if runner_coord == chaser_coord or runner_coord in self.grid.get_neighbours(
            chaser_coord, True
        ):
            r_runner, r_chaser = (self.R_CAPTURE, -self.R_CAPTURE)
        return {str(self.RUNNER_IDX): r_runner, str(self.CHASER_IDX): r_chaser}

    def _get_outcome(self, state: TPState) -> Dict[M.AgentID, M.Outcome]:
        # Assuming state is terminal
        runner_coord = state[self.RUNNER_IDX]
        chaser_coord = state[self.CHASER_IDX]
        runner_outcome, chaser_outcome = (M.Outcome.DRAW, M.Outcome.DRAW)
        if runner_coord in self.grid.goal_coords:
            runner_outcome, chaser_outcome = (M.Outcome.WIN, M.Outcome.LOSS)
        elif runner_coord == chaser_coord or runner_coord in self.grid.get_neighbours(
            chaser_coord, True
        ):
            runner_outcome, chaser_outcome = (M.Outcome.LOSS, M.Outcome.WIN)
        return {
            str(self.RUNNER_IDX): runner_outcome,
            str(self.CHASER_IDX): chaser_outcome,
        }

    def _state_is_terminal(self, state: TPState) -> bool:
        runner_coords = state[self.RUNNER_IDX]
        chaser_coords = state[self.CHASER_IDX]
        if chaser_coords == runner_coords or runner_coords in self.grid.goal_coords:
            return True
        neighbour_coords = self.grid.get_neighbours(chaser_coords, True)
        return runner_coords in neighbour_coords


class TPGrid(Grid):
    """A grid for the Two-Paths Problem."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Set[Coord],
        goal_coords: Set[Coord],
        init_runner_coord: Coord,
        init_chaser_coord: Coord,
    ):
        super().__init__(grid_width, grid_height, block_coords)
        self.goal_coords = goal_coords
        self.init_runner_coord = init_runner_coord
        self.init_chaser_coord = init_chaser_coord

    def get_ascii_repr(
        self, runner_coord: Optional[Coord], chaser_coord: Optional[Coord]
    ) -> str:
        """Get ascii repr of grid."""
        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord in self.goal_coords:
                    row_repr.append("G")
                elif coord in self.block_coords:
                    row_repr.append("#")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        if runner_coord is not None:
            grid_repr[runner_coord[1]][runner_coord[0]] = "R"
        if chaser_coord is not None:
            if chaser_coord == runner_coord:
                grid_repr[chaser_coord[1]][chaser_coord[0]] = "X"
            else:
                grid_repr[chaser_coord[1]][chaser_coord[0]] = "C"

        return "\n".join([" ".join(r) for r in grid_repr])

    def get_init_ascii_repr(self) -> str:
        """Get ascii repr of initial grid."""
        return self.get_ascii_repr(self.init_runner_coord, self.init_chaser_coord)


def get_3x3_grid() -> TPGrid:
    """Generate the Two-Paths 3-by-3 grid layout.

       012
      #####
    0 #GC #
    1 # #G#
    2 # R #
      #####

    """
    return TPGrid(
        grid_height=3,
        grid_width=3,
        block_coords={(1, 1)},
        goal_coords={(0, 0), (2, 1)},
        init_runner_coord=(1, 2),
        init_chaser_coord=(1, 0),
    )


def get_4x4_grid() -> TPGrid:
    """Generate the Choose a Path 4-by-4 grid layout.

       0123
      ######
    0 #G C #
    1 # ##G#
    2 # #  #
    3 #  R##
      ######

    """
    block_coords = {
        (1, 1),
        (2, 1),
        (1, 2),
        (3, 3),
    }

    return TPGrid(
        grid_height=4,
        grid_width=4,
        block_coords=block_coords,
        goal_coords={(0, 0), (3, 1)},
        init_runner_coord=(2, 3),
        init_chaser_coord=(2, 0),
    )


def get_7x7_grid() -> TPGrid:
    """Generate the Two-Paths 7-by-7 grid layout.

       0123456
      #########
    0 #G   C  #
    1 # ##### #
    2 # #### G#
    3 #  ### ##
    4 ## ### ##
    5 ##  #  ##
    6 ### R ###
      #########

    """
    block_coords = {
        (1, 1),
        (2, 1),
        (3, 1),
        (4, 1),
        (5, 1),
        (1, 2),
        (2, 2),
        (3, 2),
        (4, 2),
        (2, 3),
        (3, 3),
        (4, 3),
        (6, 3),
        (0, 4),
        (2, 4),
        (3, 4),
        (4, 4),
        (6, 4),
        (0, 5),
        (3, 5),
        (6, 5),
        (0, 6),
        (1, 6),
        (5, 6),
        (6, 6),
    }

    return TPGrid(
        grid_height=7,
        grid_width=7,
        block_coords=block_coords,
        goal_coords={(0, 0), (6, 2)},
        init_runner_coord=(3, 6),
        init_chaser_coord=(4, 0),
    )


# grid_size: grid_make_fn
SUPPORTED_GRIDS = {
    3: get_3x3_grid,
    4: get_4x4_grid,
    7: get_7x7_grid,
}



"""posggym specific errors.

Adapted from Farama Foundation gymnasium, copied here to so that error source path is
reported correctly so as to avoid any confusion.
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/error.py

"""


class Error(Exception):
    """Base posggym error."""


class Unregistered(Error):
    """Raised when user requests item from registry that doesn't exist."""


class UnregisteredEnv(Unregistered):
    """Raised when user requests env from registry that doesn't exist."""


class NamespaceNotFound(UnregisteredEnv):
    """Raised when user requests env from registry where namespace doesn't exist."""


class NameNotFound(UnregisteredEnv):
    """Raised when user requests env from registry where name doesn't exist."""


class VersionNotFound(UnregisteredEnv):
    """Raised when user requests env from registry where version doesn't exist."""


class DeprecatedEnv(Error):
    """Raised when user requests env from registry with old version.

    I.e. if the version number is older than the latest version env with the same
    name.
    """


class RegistrationError(Error):
    """Raised when the user attempts to register an invalid env.

    For example, an unversioned env when a versioned env exists.
    """


class UnseedableEnv(Error):
    """Raised when the user tries to seed an env that does not support seeding."""


class DependencyNotInstalled(Error):
    """Raised when the user has not installed a dependency."""


class UnsupportedMode(Error):
    """Raised when user requests rendering mode not supported by the environment."""


class InvalidMetadata(Error):
    """Raised when the metadata of an environment is not valid."""


class ResetNeeded(Error):
    """Raised when the user attempts to step environment before a reset."""


class ResetNotAllowed(Error):
    """Raised when user tries to reset an environment that's not done.

    Applicable when monitor is active.
    """


class InvalidAction(Error):
    """Raised when the user performs an action not contained within the action space."""


class MissingArgument(Error):
    """Raised when a required argument in the initializer is missing."""


class InvalidProbability(Error):
    """Raised when given an invalid value for a probability."""


class InvalidBound(Error):
    """Raised when the clipping an array with invalid upper and/or lower bound."""


# Video errors


class VideoRecorderError(Error):
    """Video recorder error."""

    pass


class InvalidFrame(Error):
    """Invalid video frame error."""

    pass



"""Setup for the posggym package."""

import pathlib

from setuptools import setup

CWD = pathlib.Path(__file__).absolute().parent


def get_version():
    """Gets the posggym version."""
    path = CWD / "posggym" / "__init__.py"
    content = path.read_text()

    for line in content.splitlines():
        if line.startswith("__version__"):
            return line.strip().split()[-1].strip().strip('"')
    raise RuntimeError("bad version data in __init__.py")


setup(name="posggym", version=get_version())



"""The classic Rock Paper Scissors problem."""
import sys
from itertools import product
from typing import Dict, List, Optional, Tuple

from gymnasium import spaces

import posggym.model as M
from posggym.core import DefaultEnv
from posggym.utils import seeding


RPSState = int
STATE0 = 0
STATES = [STATE0]
STATE_STRS = ["0"]

RPSAction = int
ROCK = 0
PAPER = 1
SCISSORS = 2
ACTIONS = [ROCK, PAPER, SCISSORS]
ACTION_STR = ["R", "P", "S"]

RPSObs = int
RPSJointObs = Tuple[RPSObs, ...]
OBS_SPACE = ACTIONS
OBS_STR = ACTION_STR


class RockPaperScissorsEnv(DefaultEnv):
    """The Rock Paper Scissors Environment.

    This is the classic game of rock, paper, scissors (RPS).

    This scenario involves two agents. Each step both agents choose an action
    out of 'ROCK', 'PAPER' or 'SCISSORS' and are rewarded based on the actions
    taken in comparison to their opponent.

    Possible Agents
    ---------------
    The environment supports two agents: '0' and '1'. Both agents are always active in
    the environment.

    State Space
    -----------
    There is only a single state in RPS: state `0`.

    Action Space
    ------------
    Both agents have three available actions: `ROCK=0`, `PAPER=1`, `SCISSORS=2`.

    Observation Space
    -----------------
    Agents observe the last action played by their opponent: `ROCK=0`, `PAPER=1`,
    `SCISSORS=2`

    Rewards
    -------
    Agents are rewarded based on the following pay-off matrix (shows pay-off
    for row agent):

    |          | ROCK     | PAPER    | SCISSORS |
    |----------|----------|----------|----------|
    | ROCK     | 0        | -1       | 1        |
    | PAPER    | 1        | 0        | -1       |
    | SCISSORS | -1       | 1        | 0        |

    Dynamics
    --------
    There is only a single state so the transition function is the identity function.

    Starting State
    --------------
    There is only a single state, which the environment always starts and remains in.

    Episode End
    -----------
    By default episodes continue infinitely long. To set a step limit, specify
    `max_episode_steps` when initializing the environment with `posggym.make`.

    Arguments
    ---------
    No additional arguments are currently supported during construction.

    Version History
    ---------------
    - `v0`: Initial version

    """

    metadata = {"render_modes": ["human", "ansi"], "render_fps": 4}

    def __init__(self, render_mode: Optional[str] = None):
        super().__init__(RockPaperScissorsModel(), render_mode=render_mode)

    def render(self, mode: str = "human"):
        assert self._last_obs is not None

        obs_str = ", ".join([OBS_STR[o] for o in self._last_obs.values()])
        output = [f"Step: {self._step_num}", f"Obs: <{obs_str}>"]
        if self._last_actions is not None:
            action_str = ", ".join([ACTION_STR[a] for a in self._last_actions.values()])
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{self._last_rewards}>")

        output_str = "\n".join(output) + "\n"

        if self.render_mode == "human":
            sys.stdout.write(output_str)
        else:
            # ansi mode
            return output_str


class RockPaperScissorsModel(M.POSGFullModel[RPSState, RPSObs, RPSAction]):
    """Rock, Paper, Scissors Model."""

    NUM_AGENTS = 2

    R_MATRIX = [[0, -1.0, 1.0], [1.0, 0, -1.0], [-1.0, 1.0, 0]]

    def __init__(self):
        self.possible_agents = tuple(str(i) for i in range(self.NUM_AGENTS))
        self.state_space = spaces.Discrete(len(STATES))
        self.action_spaces = {
            i: spaces.Discrete(len(ACTIONS)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: spaces.Discrete(len(OBS_SPACE)) for i in self.possible_agents
        }

        # Spaces used internally
        self._state_space = STATES
        self._action_spaces = tuple([*ACTIONS] for _ in self.possible_agents)
        self._obs_spaces = tuple(OBS_SPACE for _ in self.possible_agents)
        self.is_symmetric = True

        self._trans_map = self._construct_trans_func()
        self._rew_map = self._construct_rew_func()
        self._obs_map = self._construct_obs_func()

    @property
    def reward_ranges(self) -> Dict[M.AgentID, Tuple[float, float]]:
        return {i: (-1.0, 1.0) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: RPSState) -> List[M.AgentID]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> RPSState:
        return STATE0

    def sample_initial_obs(self, state: RPSState) -> Dict[M.AgentID, RPSObs]:
        return {i: ROCK for i in self.possible_agents}

    def step(
        self, state: RPSState, actions: Dict[M.AgentID, RPSAction]
    ) -> M.JointTimestep[RPSState, RPSObs]:
        assert all(a_i in ACTIONS for a_i in actions.values())
        obs: Dict[M.AgentID, RPSObs] = {"0": actions["1"], "1": actions["0"]}
        rewards = self._get_reward(actions)
        terminated = {i: False for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        all_done = False
        info: Dict[M.AgentID, Dict] = {i: {} for i in self.possible_agents}
        return M.JointTimestep(
            STATE0, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_reward(
        self, actions: Dict[M.AgentID, RPSAction]
    ) -> Dict[M.AgentID, float]:
        return {
            "0": self.R_MATRIX[actions["0"]][actions["1"]],
            "1": self.R_MATRIX[actions["1"]][actions["0"]],
        }

    def get_initial_belief(self) -> Dict[RPSState, float]:
        return {STATE0: 1.0}

    def transition_fn(
        self, state: RPSState, actions: Dict[M.AgentID, RPSAction], next_state: RPSState
    ) -> float:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._trans_map[(state, action_tuple, next_state)]

    def _construct_trans_func(self) -> Dict:
        trans_map = {}
        for a in product(*self._action_spaces):
            trans_map[(STATE0, a, STATE0)] = 1.0
        return trans_map

    def observation_fn(
        self,
        obs: Dict[M.AgentID, RPSObs],
        next_state: RPSState,
        actions: Dict[M.AgentID, RPSAction],
    ) -> float:
        obs_tuple = tuple(obs[i] for i in self.possible_agents)
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._obs_map[(next_state, action_tuple, obs_tuple)]

    def _construct_obs_func(self) -> Dict:
        obs_func = {}
        for a, o in product(product(*self._action_spaces), product(*self._obs_spaces)):
            obs_func[(STATE0, a, o)] = 1.0 if a == o else 0.0
        return obs_func

    def reward_fn(
        self, state: RPSState, actions: Dict[M.AgentID, RPSAction]
    ) -> Dict[M.AgentID, float]:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._rew_map[(state, action_tuple)]

    def _construct_rew_func(self) -> Dict:
        rew_map = {}
        for a in product(*self._action_spaces):
            rew_map[(STATE0, a)] = {
                "0": self.R_MATRIX[a[0]][a[1]],
                "1": self.R_MATRIX[a[1]][a[0]],
            }
        return rew_map
